# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum
from column_name import ColumnName


class MetadataModelData:

    def __init__(
            self,
            column_name: ColumnName,
            primary_key: bool = False,
            calculated: bool = False,
            is_created_timestamp: bool = False,
            is_updated_timestamp: bool = False):
        self.__column_name: ColumnName = column_name
        self.__primary_key: bool = primary_key
        self.__calculated: bool = calculated
        self.__is_created_timestamp: bool = is_created_timestamp
        self.__is_updated_timestamp: bool = is_updated_timestamp

    @property
    def column_name(self) -> ColumnName:
        return self.__column_name

    @property
    def primary_key(self) -> bool:
        return self.__primary_key

    @property
    def calculated(self) -> bool:
        return self.__calculated

    @property
    def is_created_timestamp(self) -> bool:
        return self.__is_created_timestamp

    @property
    def is_updated_timestamp(self) -> bool:
        return self.__is_updated_timestamp


class ArtistMetadataModel(Enum):

    ARTIST_ID = MetadataModelData(
        column_name=ColumnName.ARTIST_ID,
        primary_key=True)
    ARTIST_NAME = MetadataModelData(column_name=ColumnName.ARTIST_NAME)
    ARTIST_MB_ID = MetadataModelData(column_name=ColumnName.ARTIST_MB_ID)
    ARTIST_ALBUM_COUNT = MetadataModelData(column_name=ColumnName.ARTIST_ALBUM_COUNT)
    ARTIST_COVER_ART = MetadataModelData(column_name=ColumnName.ARTIST_COVER_ART)
    ARTIST_MEDIA_TYPE = MetadataModelData(column_name=ColumnName.ARTIST_MEDIA_TYPE)
    ARTIST_SORT_NAME = MetadataModelData(column_name=ColumnName.ARTIST_SORT_NAME)
    ARTIST_STARRED = MetadataModelData(column_name=ColumnName.ARTIST_STARRED)
    CREATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.CREATED_TIMESTAMP,
        is_created_timestamp=True)
    UPDATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.UPDATED_TIMESTAMP,
        is_updated_timestamp=True)

    @property
    def column_name(self) -> ColumnName:
        return self.value.column_name

    @property
    def primary_key(self) -> bool:
        return self.value.primary_key

    @property
    def calculated(self) -> bool:
        return self.value.calculated

    @property
    def is_created_timestamp(self) -> bool:
        return self.value.is_created_timestamp

    @property
    def is_updated_timestamp(self) -> bool:
        return self.value.is_updated_timestamp


class AlbumMetadataModel(Enum):

    ALBUM_ID = MetadataModelData(column_name=ColumnName.ALBUM_ID, primary_key=True)
    QUALITY_BADGE = MetadataModelData(
        column_name=ColumnName.QUALITY_BADGE,
        calculated=True)
    ALBUM_MB_ID = MetadataModelData(column_name=ColumnName.ALBUM_MB_ID)
    ALBUM_ARTIST_ID = MetadataModelData(column_name=ColumnName.ALBUM_ARTIST_ID)
    ALBUM_PATH = MetadataModelData(column_name=ColumnName.ALBUM_PATH, calculated=True)
    ALBUM_ARTIST = MetadataModelData(column_name=ColumnName.ALBUM_ARTIST)
    ALBUM_NAME = MetadataModelData(column_name=ColumnName.ALBUM_NAME)
    ALBUM_COVER_ART = MetadataModelData(column_name=ColumnName.ALBUM_COVER_ART)
    ALBUM_DISC_COUNT = MetadataModelData(column_name=ColumnName.ALBUM_DISC_COUNT)
    ALBUM_SONG_COUNT = MetadataModelData(column_name=ColumnName.ALBUM_SONG_COUNT)
    ALBUM_DURATION = MetadataModelData(column_name=ColumnName.ALBUM_DURATION)
    ALBUM_CREATED = MetadataModelData(column_name=ColumnName.ALBUM_CREATED)
    ALBUM_YEAR = MetadataModelData(column_name=ColumnName.ALBUM_YEAR)
    ALBUM_GENRE = MetadataModelData(column_name=ColumnName.ALBUM_GENRE)
    ALBUM_GENRE_LIST = MetadataModelData(column_name=ColumnName.ALBUM_GENRE_LIST)
    ALBUM_USER_RATING = MetadataModelData(column_name=ColumnName.ALBUM_USER_RATING)
    ALBUM_DISPLAY_ARTIST = MetadataModelData(column_name=ColumnName.ALBUM_DISPLAY_ARTIST)
    ALBUM_EXPLICIT_STATUS = MetadataModelData(column_name=ColumnName.EXPLICIT_STATUS)
    ALBUM_IS_COMPILATION = MetadataModelData(column_name=ColumnName.ALBUM_IS_COMPILATION)
    ALBUM_PLAY_COUNT = MetadataModelData(column_name=ColumnName.ALBUM_PLAY_COUNT)
    ALBUM_PLAYED = MetadataModelData(column_name=ColumnName.ALBUM_PLAYED)
    ALBUM_SORT_NAME = MetadataModelData(column_name=ColumnName.ALBUM_SORT_NAME)
    ALBUM_VERSION = MetadataModelData(column_name=ColumnName.ALBUM_VERSION)
    ALBUM_ORIGINAL_RELEASE_DATE_YEAR = MetadataModelData(column_name=ColumnName.ALBUM_ORIGINAL_RELEASE_DATE_YEAR)
    ALBUM_ORIGINAL_RELEASE_DATE_MONTH = MetadataModelData(column_name=ColumnName.ALBUM_ORIGINAL_RELEASE_DATE_MONTH)
    ALBUM_ORIGINAL_RELEASE_DATE_DAY = MetadataModelData(column_name=ColumnName.ALBUM_ORIGINAL_RELEASE_DATE_DAY)
    ALBUM_ORIGINAL_RELEASE_DATE_STR = MetadataModelData(column_name=ColumnName.ALBUM_ORIGINAL_RELEASE_DATE_STR, calculated=True)
    ALBUM_RELEASE_DATE_YEAR = MetadataModelData(column_name=ColumnName.ALBUM_RELEASE_DATE_YEAR)
    ALBUM_RELEASE_DATE_MONTH = MetadataModelData(column_name=ColumnName.ALBUM_RELEASE_DATE_MONTH)
    ALBUM_RELEASE_DATE_DAY = MetadataModelData(column_name=ColumnName.ALBUM_RELEASE_DATE_DAY)
    ALBUM_RELEASE_DATE_STR = MetadataModelData(column_name=ColumnName.ALBUM_RELEASE_DATE_STR, calculated=True)
    ALBUM_TRACK_QUALITY_SUMMARY = MetadataModelData(
        column_name=ColumnName.ALBUM_TRACK_QUALITY_SUMMARY,
        calculated=True)
    ALBUM_AVERAGE_BITRATE = MetadataModelData(
        column_name=ColumnName.ALBUM_AVERAGE_BITRATE,
        calculated=True)
    ALBUM_LOSSLESS_STATUS = MetadataModelData(
        column_name=ColumnName.ALBUM_LOSSLESS_STATUS,
        calculated=True)
    ALBUM_MOOD_LIST = MetadataModelData(column_name=ColumnName.ALBUM_MOOD_LIST)
    ALBUM_RECORD_LABEL_LIST = MetadataModelData(column_name=ColumnName.ALBUM_RECORD_LABEL_LIST)
    ALBUM_RELEASE_TYPE_LIST = MetadataModelData(column_name=ColumnName.ALBUM_RELEASE_TYPE_LIST)
    ALBUM_MEDIA_TYPE = MetadataModelData(column_name=ColumnName.ALBUM_MEDIA_TYPE)
    ALBUM_STARRED = MetadataModelData(column_name=ColumnName.ALBUM_STARRED)
    CREATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.CREATED_TIMESTAMP,
        is_created_timestamp=True)
    UPDATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.UPDATED_TIMESTAMP,
        is_updated_timestamp=True)

    @property
    def column_name(self) -> ColumnName:
        return self.value.column_name

    @property
    def primary_key(self) -> bool:
        return self.value.primary_key

    @property
    def calculated(self) -> bool:
        return self.value.calculated

    @property
    def is_created_timestamp(self) -> bool:
        return self.value.is_created_timestamp

    @property
    def is_updated_timestamp(self) -> bool:
        return self.value.is_updated_timestamp


class SongMetadataModel(Enum):

    SONG_ID = MetadataModelData(column_name=ColumnName.SONG_ID, primary_key=True)
    SONG_TITLE = MetadataModelData(column_name=ColumnName.SONG_TITLE)
    SONG_ALBUM_ID = MetadataModelData(column_name=ColumnName.SONG_ALBUM_ID)
    SONG_ARTIST_ID = MetadataModelData(column_name=ColumnName.SONG_ARTIST_ID)
    SONG_ARTIST = MetadataModelData(column_name=ColumnName.SONG_ARTIST)
    SONG_COMMENT = MetadataModelData(column_name=ColumnName.SONG_COMMENT)
    SONG_BITDEPTH = MetadataModelData(column_name=ColumnName.SONG_BITDEPTH)
    SONG_BITRATE = MetadataModelData(column_name=ColumnName.SONG_BITRATE)
    SONG_CHANNEL_COUNT = MetadataModelData(column_name=ColumnName.SONG_CHANNEL_COUNT)
    SONG_SAMPLING_RATE = MetadataModelData(column_name=ColumnName.SONG_SAMPLING_RATE)
    SONG_SIZE = MetadataModelData(column_name=ColumnName.SONG_SIZE)
    SONG_SUFFIX = MetadataModelData(column_name=ColumnName.SONG_SUFFIX)
    SONG_DISC_NUMBER = MetadataModelData(column_name=ColumnName.SONG_DISC_NUMBER)
    SONG_TRACK = MetadataModelData(column_name=ColumnName.SONG_TRACK)
    SONG_TYPE = MetadataModelData(column_name=ColumnName.SONG_TYPE)
    SONG_DURATION = MetadataModelData(column_name=ColumnName.SONG_DURATION)
    SONG_CONTENT_TYPE = MetadataModelData(column_name=ColumnName.SONG_CONTENT_TYPE)
    SONG_COVER_ART = MetadataModelData(column_name=ColumnName.SONG_COVER_ART)
    SONG_CREATED = MetadataModelData(column_name=ColumnName.SONG_CREATED)
    SONG_DISPLAY_ALBUM_ARTIST = MetadataModelData(column_name=ColumnName.SONG_DISPLAY_ALBUM_ARTIST)
    SONG_DISPLAY_ARTIST = MetadataModelData(column_name=ColumnName.SONG_DISPLAY_ARTIST)
    SONG_EXPLICIT_STATUS = MetadataModelData(column_name=ColumnName.SONG_EXPLICIT_STATUS)
    SONG_GENRE = MetadataModelData(column_name=ColumnName.SONG_GENRE)
    SONG_IS_DIR = MetadataModelData(column_name=ColumnName.SONG_IS_DIR)
    SONG_MEDIA_TYPE = MetadataModelData(column_name=ColumnName.SONG_MEDIA_TYPE)
    SONG_MUSICBRAINZ_ID = MetadataModelData(column_name=ColumnName.SONG_MUSICBRAINZ_ID)
    SONG_PATH = MetadataModelData(column_name=ColumnName.SONG_PATH)
    SONG_PLAY_COUNT = MetadataModelData(column_name=ColumnName.SONG_PLAY_COUNT)
    SONG_PLAYED = MetadataModelData(column_name=ColumnName.SONG_PLAYED)
    SONG_YEAR = MetadataModelData(column_name=ColumnName.SONG_YEAR)
    SONG_DISPLAY_COMPOSER = MetadataModelData(column_name=ColumnName.SONG_DISPLAY_COMPOSER)
    SONG_SORT_NAME = MetadataModelData(column_name=ColumnName.SONG_SORT_NAME)
    SONG_LOSSLESS_STATUS = MetadataModelData(column_name=ColumnName.SONG_LOSSLESS_STATUS)
    SONG_STARRED = MetadataModelData(column_name=ColumnName.SONG_STARRED)
    CREATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.CREATED_TIMESTAMP,
        is_created_timestamp=True)
    UPDATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.UPDATED_TIMESTAMP,
        is_updated_timestamp=True)

    @property
    def column_name(self) -> ColumnName:
        return self.value.column_name

    @property
    def primary_key(self) -> bool:
        return self.value.primary_key

    @property
    def calculated(self) -> bool:
        return self.value.calculated

    @property
    def is_created_timestamp(self) -> bool:
        return self.value.is_created_timestamp

    @property
    def is_updated_timestamp(self) -> bool:
        return self.value.is_updated_timestamp


class AlbumArtistMetaModel(Enum):

    ID = MetadataModelData(column_name=ColumnName.ID, primary_key=True)
    ALBUM_ID = MetadataModelData(column_name=ColumnName.ALBUM_ID)
    ARTIST_ID = MetadataModelData(column_name=ColumnName.ARTIST_ID)
    ARTIST_NAME = MetadataModelData(column_name=ColumnName.ARTIST_NAME)
    CREATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.CREATED_TIMESTAMP,
        is_created_timestamp=True)

    @property
    def column_name(self) -> ColumnName:
        return self.value.column_name

    @property
    def is_created_timestamp(self) -> bool:
        return self.value.is_created_timestamp


class SongAlbumArtistMetaModel(Enum):

    ID = MetadataModelData(column_name=ColumnName.ID, primary_key=True)
    SONG_ID = MetadataModelData(column_name=ColumnName.SONG_ID)
    SONG_ALBUM_ID = MetadataModelData(column_name=ColumnName.SONG_ALBUM_ID)
    SONG_ALBUM_ARTIST_ID = MetadataModelData(column_name=ColumnName.SONG_ALBUM_ARTIST_ID)
    CREATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.CREATED_TIMESTAMP,
        is_created_timestamp=True)

    @property
    def column_name(self) -> ColumnName:
        return self.value.column_name

    @property
    def calculated(self) -> bool:
        return self.value.calculated

    @property
    def is_created_timestamp(self) -> bool:
        return self.value.is_created_timestamp

    @property
    def is_updated_timestamp(self) -> bool:
        return self.value.is_updated_timestamp


class SongArtistMetaModel(Enum):

    ID = MetadataModelData(column_name=ColumnName.ID, primary_key=True)
    SONG_ID = MetadataModelData(column_name=ColumnName.SONG_ID)
    SONG_ALBUM_ID = MetadataModelData(column_name=ColumnName.SONG_ALBUM_ID)
    SONG_ARTIST_ID = MetadataModelData(column_name=ColumnName.SONG_ARTIST_ID)
    CREATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.CREATED_TIMESTAMP,
        is_created_timestamp=True)

    @property
    def column_name(self) -> ColumnName:
        return self.value.column_name

    @property
    def calculated(self) -> bool:
        return self.value.calculated

    @property
    def is_created_timestamp(self) -> bool:
        return self.value.is_created_timestamp

    @property
    def is_updated_timestamp(self) -> bool:
        return self.value.is_updated_timestamp


class SongContributorMetaModel(Enum):

    ID = MetadataModelData(column_name=ColumnName.ID)
    SONG_ID = MetadataModelData(column_name=ColumnName.SONG_ID)
    SONG_ALBUM_ID = MetadataModelData(column_name=ColumnName.SONG_ALBUM_ID)
    SONG_CONTRIBUTOR_ROLE = MetadataModelData(column_name=ColumnName.SONG_CONTRIBUTOR_ROLE)
    SONG_CONTRIBUTOR_SUB_ROLE = MetadataModelData(column_name=ColumnName.SONG_CONTRIBUTOR_SUB_ROLE)
    SONG_ARTIST_ID = MetadataModelData(column_name=ColumnName.SONG_ARTIST_ID)
    CREATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.CREATED_TIMESTAMP,
        is_created_timestamp=True)

    @property
    def column_name(self) -> ColumnName:
        return self.value.column_name

    @property
    def calculated(self) -> bool:
        return self.value.calculated

    @property
    def is_created_timestamp(self) -> bool:
        return self.value.is_created_timestamp

    @property
    def is_updated_timestamp(self) -> bool:
        return self.value.is_updated_timestamp


class AlbumPropertyMetaModel(Enum):

    ALBUM_ID = MetadataModelData(column_name=ColumnName.ALBUM_ID)
    ALBUM_PROPERTY_KEY = MetadataModelData(column_name=ColumnName.ALBUM_PROPERTY_KEY)
    ALBUM_PROPERTY_VALUE = MetadataModelData(column_name=ColumnName.ALBUM_PROPERTY_VALUE)
    CREATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.CREATED_TIMESTAMP,
        is_created_timestamp=True)
    UPDATED_TIMESTAMP = MetadataModelData(
        column_name=ColumnName.UPDATED_TIMESTAMP,
        is_updated_timestamp=True)

    @property
    def column_name(self) -> ColumnName:
        return self.value.column_name

    @property
    def calculated(self) -> bool:
        return self.value.calculated

    @property
    def is_created_timestamp(self) -> bool:
        return self.value.is_created_timestamp

    @property
    def is_updated_timestamp(self) -> bool:
        return self.value.is_updated_timestamp



# Copyright (C) 2016-2022 J.F.Dockes
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# This implements a cmdtalk processor (class Processor) which expects a method name in each set of
# input parameters, under the key prcnmkey: "cmdtalk:proc". The client module creates a Dispatch
# object which it uses to record the methods to call. The message processor looks for the prcnmkey
# value and call the processing method accordingly. Example code:
#
#     dispatcher = cmdtalkplugin.Dispatch()
#     msgproc = cmdtalkplugin.Processor(dispatcher)
#
#     @dispatcher.record('trackuri')
#     def trackuri(a):
#       dosomething(a)
#     ....
#
#     msgproc.mainloop()

import sys
import cmdtalk

prcnmkey = "cmdtalk:proc"


class Dispatch:
    def __init__(self):
        self.map = {}

    def record(self, nm):
        def decorator(func):
            self.map[nm] = func
            return func

        return decorator

    def run(self, nm, params):
        func = self.map[nm]
        return func(params)


class Processor:
    def __init__(self, dispatcher, outfile=sys.stdout, infile=sys.stdin, exitfunc=None):
        self.em = cmdtalk.CmdTalk(outfile=outfile, infile=infile, exitfunc=exitfunc)
        self.dispatcher = dispatcher

    def log(self, s, doexit=0, exitvalue=1):
        self.em.log(s, doexit, exitvalue)

    def process(self, params):
        # self.em.log(f"pCmdTalkProcessor.process: [{params}]")
        if not prcnmkey in params:
            raise Exception(f"{prcnmkey} not in args")

        return self.dispatcher.run(params[prcnmkey], params)

    def mainloop(self):
        cmdtalk.main(self.em, self)



# Copyright (C) 2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# this should contain all methods which interact directly with the subsonic server

from subsonic_connector.response import Response
from subsonic_connector.album_list import AlbumList
from subsonic_connector.artists import Artists
from subsonic_connector.genres import Genres
from subsonic_connector.starred import Starred
from subsonic_connector.playlists import Playlists
from subsonic_connector.artist import Artist
from subsonic_connector.search_result import SearchResult
from subsonic_connector.list_type import ListType
from msgproc_provider import msgproc
import config
import constants
import connector_provider
import datetime
import time


class CachedResponse:
    last_response_obj: Response = None
    last_response_time: datetime.datetime = None


class CachedArtistList:
    last_Artist_List: list[Artist] = None
    last_response_time: datetime.datetime = None


cached_response_artists: CachedResponse = CachedResponse()
cached_response_random_firstpage: CachedResponse = CachedResponse()
cached_response_genres: CachedResponse = CachedResponse()
cached_starred_albums: CachedResponse = CachedResponse()
cached_starred: CachedResponse = CachedResponse()
cached_playlists: CachedResponse = CachedResponse()
cached_response_newest: CachedResponse = CachedResponse()

cached_all_artist_list: CachedArtistList = CachedArtistList()


def __is_older_than(date_time: datetime.datetime, delta_sec: int) -> bool:
    if date_time is None:
        return True
    cutoff: datetime.datetime = datetime.datetime.now() - datetime.timedelta(seconds=delta_sec)
    return date_time < cutoff


def __cached_list_is_expired(cached: CachedArtistList, delta_sec: int) -> bool:
    return (cached is None or
            cached.last_Artist_List is None or
            __is_older_than(
                cached.last_response_time
                if cached is not None
                else None,
                delta_sec))


def __cached_response_is_expired(cached: CachedResponse, delta_sec: int) -> bool:
    return (cached is None or
            cached.last_response_obj is None or
            __is_older_than(
                cached.last_response_time
                if cached is not None
                else None,
                delta_sec))


def get_playlists() -> Response[Artists]:
    global cached_playlists
    if __cached_response_is_expired(
            cached=cached_playlists,
            delta_sec=config.get_config_param_as_int(constants.ConfigParam.CACHED_REQUEST_TIMEOUT_SEC)):
        msgproc.log("subsonic_util.get_playlists loading playlists ...")
        # actually request starred
        res: Response[Playlists] = connector_provider.get().getPlaylists()
        # store response along with timestamp
        cached_playlists = CachedResponse()
        cached_playlists.last_response_obj = res
        cached_playlists.last_response_time = datetime.datetime.now()
        return res
    else:
        return cached_playlists.last_response_obj


def get_starred() -> Response[Starred]:
    verbose: bool = config.get_verbose_logging()
    global cached_starred
    if __cached_response_is_expired(
            cached=cached_starred,
            delta_sec=config.get_config_param_as_int(constants.ConfigParam.CACHED_REQUEST_TIMEOUT_SEC)):
        msgproc.log("subsonic_util.get_starred loading starred ...")
        # actually request starred
        start: float = time.time()
        res: Response[Starred] = connector_provider.get().getStarred()
        # store response along with timestamp
        cached_starred = CachedResponse()
        cached_starred.last_response_obj = res
        cached_starred.last_response_time = datetime.datetime.now()
        if verbose:
            msgproc.log(f"request_cache.get_starred took [{(time.time() - start):.3f}] sec")
        return res
    else:
        if verbose:
            msgproc.log("request_cache.get_starred is using cached data")
        return cached_starred.last_response_obj


def get_all_artists() -> list[Artist]:
    global cached_all_artist_list
    if __cached_list_is_expired(
            cached=cached_all_artist_list,
            delta_sec=config.get_config_param_as_int(constants.ConfigParam.CACHED_ARTIST_LIST_CACHE_TIMEOUT_SEC)):
        msgproc.log("subsonic_util.get_all_artists loading all artists ...")
        all_artists: list[Artist] = _load_all_artists()
        cached_all_artist_list = CachedArtistList()
        cached_all_artist_list.last_Artist_List = all_artists
        cached_all_artist_list.last_response_time = datetime.datetime.now()
        return all_artists
    else:
        return cached_all_artist_list.last_Artist_List


def _load_all_artists() -> list[Artist]:
    all_artist: list[Artist] = []
    finished: bool = False
    search_size: int = 1000
    search_offset: int = 0
    while not finished:
        msgproc.log(f"Executing search with offset [{search_offset}]")
        search_result: SearchResult = connector_provider.get().search(
            "",
            artistCount=search_size,
            songCount=0,
            albumCount=0,
            artistOffset=search_offset)
        artists: list[Artist] = search_result.getArtists()
        all_artist.extend(artists)
        ac: int = len(artists)
        search_offset += ac
        if ac < search_size:
            finished = True
    msgproc.log(f"Sorting [{len(all_artist)}] artists ...")
    return all_artist


def get_artists() -> Response[Artists]:
    global cached_response_artists
    if __cached_response_is_expired(
            cached=cached_response_artists,
            delta_sec=config.get_config_param_as_int(constants.ConfigParam.CACHED_REQUEST_TIMEOUT_SEC)):
        msgproc.log("subsonic_util.get_artists loading artists ...")
        # actually request artists
        res: Response[Artists] = connector_provider.get().getArtists()
        # store response along with timestamp
        cached_response_artists = CachedResponse()
        cached_response_artists.last_response_obj = res
        cached_response_artists.last_response_time = datetime.datetime.now()
        msgproc.log("subsonic_util.get_artists artists have been loaded.")
        return res
    else:
        return cached_response_artists.last_response_obj


def get_random_album_list(
        size=10,
        offset=0,
        fromYear=None,
        toYear=None,
        genre=None,
        musicFolderId=None) -> Response[AlbumList]:
    # cache first page
    if not (
            size == config.get_items_per_page() and
            int(offset) == 0 and
            genre is None and
            fromYear is None and
            toYear is None and
            musicFolderId is None):
        msgproc.log(f"request_cache.get_random_album_list cannot be cached "
                    f"size:[{size}] [{size == config.get_items_per_page()}] "
                    f"offset:[{offset}] [{int(offset) == 0}] "
                    f"fromYear:[{fromYear}] [{fromYear is None}] "
                    f"toYear:[{toYear}] [{toYear is None}] "
                    f"genre:[{genre}] [{genre is None}] "
                    f"musicFolderId:[{musicFolderId}] [{musicFolderId is None}]")
        return connector_provider.get().getRandomAlbumList(
            size=size,
            offset=offset,
            fromYear=fromYear,
            toYear=toYear,
            genre=genre,
            musicFolderId=musicFolderId)
    else:
        global cached_response_random_firstpage
        if __cached_response_is_expired(
                cached=cached_response_random_firstpage,
                delta_sec=config.get_config_param_as_int(constants.ConfigParam.CACHED_REQUEST_TIMEOUT_SEC)):
            msgproc.log(f"subsonic_util.get_random_album_list loading first [{config.get_items_per_page()}] random albums ...")
            # actually request first random albums
            res: Response[AlbumList] = connector_provider.get().getRandomAlbumList(size=config.get_items_per_page())
            cached_response_random_firstpage = CachedResponse()
            cached_response_random_firstpage.last_response_obj = res
            cached_response_random_firstpage.last_response_time = datetime.datetime.now()
            return res
        else:
            # use cached!
            msgproc.log(f"subsonic_util.get_random_album_list returning first [{config.get_items_per_page()}] using cached data.")
            return cached_response_random_firstpage.last_response_obj


def get_genres() -> Response[Genres]:
    global cached_response_genres
    if __cached_response_is_expired(
            cached=cached_response_genres,
            delta_sec=config.get_config_param_as_int(constants.ConfigParam.CACHED_REQUEST_TIMEOUT_SEC)):
        msgproc.log("subsonic_util.get_genres loading ...")
        # actually request genres
        res: Response[Genres] = connector_provider.get().getGenres()
        cached_response_genres = CachedResponse()
        cached_response_genres.last_response_obj = res
        cached_response_genres.last_response_time = datetime.datetime.now()
        msgproc.log(f"subsonic_util.get_genres finished loading [{len(res.getObj().getGenres())}] genres")
        return res
    else:
        # use cached!
        return cached_response_genres.last_response_obj


def get_first_newest_album_list() -> list[AlbumList]:
    global cached_response_newest
    if __cached_response_is_expired(
            cached=cached_response_newest,
            delta_sec=config.get_config_param_as_int(constants.ConfigParam.CACHED_REQUEST_TIMEOUT_SEC)):
        msgproc.log("subsonic_util.get_first_newest_album_list loading ...")
        # actually request genres
        res: Response[AlbumList] = connector_provider.get().getAlbumList(
            ltype=ListType.BY_YEAR,
            size=config.get_items_per_page(),
            fromYear=datetime.datetime.now().year,
            toYear=0)
        cached_response_newest = CachedResponse()
        cached_response_newest.last_response_obj = res
        cached_response_newest.last_response_time = datetime.datetime.now()
        msgproc.log(f"subsonic_util.get_first_newest_album_list finished loading [{len(res.getObj().getAlbums())}] albums")
        return res
    else:
        # use cached!
        msgproc.log("subsonic_util.get_first_newest_album_list using cached response")
        return cached_response_newest.last_response_obj



#
# Copyright (C) 2017-2019 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function

import sys
from urllib.parse import quote as urlquote, unquote_to_bytes as urlunquotetobytes
import functools
import glob
import io
import locale
import mutagen
import os
import re
import traceback
import xml.sax.saxutils

from upmplgutils import uplog

_g_fse = sys.getfilesystemencoding()

audiomtypes = frozenset(
    [
        "audio/mpeg",
        "audio/flac",
        "application/flac",
        "audio/x-dsf",
        "audio/x-dff",
        "audio/x-aiff",
        "audio/x-flac",
        "application/x-flac",
        "application/ogg",
        "audio/aac",
        "audio/mp4",
        "video/mp4",
        "audio/x-aiff",
        "audio/x-musepack",
        "audio/ape",
        "audio/x-wav",
        "audio/x-wavpack",
        "inode/directory",
        "audio/x-mpegurl",
    ]
)

# Correspondance between Recoll field names (on the right), defined by rclaudio and the Recoll
# configuration 'fields' file, and what plgwithslave.cxx expects, which is less than consistent.
# Some tags have no predefined fields in upmpdcli (e.g. composer, conductor), and we output them as
# a didl fragment (see below). They are still listed here because the data is also used to define
# the fields actually extracted in the results of the initial recoll listing (resultstore)
upnp2rclfields = {
    "upnp:album": "album",
    "upnp:albumArtURI": "albumarturi",
    "upnp:artist": "artist",
    "dc:description": "comment",
    "composer": "composer",
    "conductor": "conductor",
    "dc:date": "date",
    "upnp:genre": "genre",
    "duration": "duration",  # should be res:
    "res:bitrate": "bitrate",
    "res:bitsPerSample": "bits_per_sample",
    "res:channels": "channels",
    "res:mime": "mtype",
    "res:samplefreq": "sample_rate",
    "res:size": "fbytes",
    "tt": "title",
    "dc:title": "title",
    "upnp:originalTrackNumber": "tracknumber",
}

_g_folders = None
def importfolders(folders):
    global _g_folders
    _g_folders = folders

def httpurl(httphp, path, query=""):
    """Compose an http URL from host path (http://xxx/) path and query.
    path can be either str or bytes and will be url-encoded"""
    return f"http://{httphp}{urlquote(path)}{query}"


def rcldoctoentry(id, pid, httphp, pathprefix, doc):
    """
    Transform a Doc object into the format expected by the parent

    Args:
        id (str): objid for the entry
        pid (str):  objid for the browsed object (the parent container)
        httphp: the hostport part of the generated track urls
        pathprefix: is provided by our parent process (it's used to
          what plugin an url belongs too when needed for
          translating the internal into the real url (for plugins
          based on external-services)
        doc is the Doc object to be translated

    Returns:
        A dict representing an UPnP item, with the
        keys as expected in the plgwithslave.cxx resultToEntries() function.
    """
    # uplog("rcldoctoentry:  pid %s id %s mtype %s" % (pid, id, doc["mtype"]))

    li = {}
    if doc["mtype"] not in audiomtypes:
        return li

    li["pid"] = pid
    li["id"] = id
    if doc["mtype"] == "inode/directory":
        li["tp"] = "ct"
        li["upnp:class"] = "object.container"
    else:
        li["tp"] = "it"
        # TBD
        li["upnp:class"] = "object.item.audioItem.musicTrack"

    for oname, dname in upnp2rclfields.items():
        val = doc[dname]
        if val:
            li[oname] = val

    if "upnp:albumArtURI" not in li:
        arturi = _g_folders.docarturi(doc, False, doc["album"])
        if arturi:
            li["upnp:albumArtURI"] = arturi

    if "upnp:artist" not in li and doc["albumartist"]:
        li["upnp:artist"] = doc["albumartist"]

    # TBD Date format ?
    # discnumber=
    # genre=
    # lyricist=
    # lyrics=

    # Some properties are not directly supported by the upmpdcli internal UpSong object, which only
    # manages basic UPnP props. We send them directly as a DIDL fragment which will be incorporated
    # in the final XML
    didlfrag = ""
    if doc["composer"]:
        didlfrag += (
            '<upnp:artist role="Composer">'
            + xml.sax.saxutils.escape(doc["composer"])
            + "</upnp:artist>"
        )
    if doc["conductor"]:
        didlfrag += (
            '<upnp:artist role="Conductor">'
            + xml.sax.saxutils.escape(doc["conductor"])
            + "</upnp:artist>"
        )
    if doc["albumartist"]:
        didlfrag += (
            '<upnp:artist role="AlbumArtist">'
            + xml.sax.saxutils.escape(doc["albumartist"])
            + "</upnp:artist>"
        )
    if didlfrag:
        li["didlfrag"] = didlfrag

    try:
        val = li["upnp:originalTrackNumber"]
        l = val.split("/")
        li["upnp:originalTrackNumber"] = l[0]
    except:
        pass

    # Compute the url. We use the URL from recoll, stripped of file://
    # and with the pathprefix prepended (the pathprefix is used by our
    # parent process to match urls to plugins)
    url = doc["url"]
    ssidx = url.find("//")
    if url.find("file://") == 0:
        bpath = docbinpath(doc)
        li["uri"] = httpurl(httphp, bpath)
    else:
        # hopefully an ok http url
        li["uri"] = url
    # uplog("rcldoctoentry: uri: %s" % li['uri'])

    if "tt" not in li:
        li["tt"] = os.path.basename(url[ssidx + 2 :])

    return li


# Bogus entry for the top directory while the index/db is updating
def waitentry(id, pid, httphp, msg="Initializing..."):
    li = {}
    li["tp"] = "it"
    li["id"] = id
    li["pid"] = pid
    li["upnp:class"] = "object.item.audioItem.musicTrack"
    li["tt"] = msg
    li["uri"] = "http://%s%s" % (httphp, "/waiting")
    li["res.mime"] = "audio/mpeg"
    return li


# dirname and basename which returns the last element, not null, when
# the path ends in '/'
def dirname(path):
    if path[-1] == "/"[0] and path != "/":
        return os.path.dirname(path[0:-1])
    else:
        return os.path.dirname(path)


def basename(path):
    if path[-1] == "/"[0] and path != "/":
        return os.path.basename(path[0:-1])
    else:
        return os.path.basename(path)


# Compute binary fs path for URL. All Recoll URLs are like file://xx
def docbinpath(doc):
    # Versions of recoll from 1.43.11 decode the binary URL to UTF-8 with the surrogateescape error
    # handler. We don't support older version any more. Not sure how we do with non-utf-8 paths
    # (TBD) For reference, minim does not process them at all.
    return doc["url"][7:].encode(_g_fse, "surrogateescape")
def docpath(doc):
    # Versions of recoll from 1.43.11 decode the binary URL to UTF-8 with the surrogateescape error
    # handler. We don't support older version any more. Not sure how we do with non-utf-8 paths
    # (TBD) For reference, minim does not process them at all.
    return doc["url"][7:]


def docfolder(doc):
    path = doc["url"][7:]
    if doc["mtype"] != "inode/directory":
        path = os.path.dirname(path)
    if path[-1] != "/"[0]:
        path += "/"
    return path


def embdimgurl(doc, httphp, pathprefix):
    if isinstance(pathprefix, str):
        pathprefix = pathprefix.encode("UTF-8")
    if doc["embdimg"] == "jpg":
        ext = b".jpg"
    elif doc["embdimg"] == "png":
        ext = b".png"
    else:
        return None
    path = docbinpath(doc)
    path = os.path.join(pathprefix, path + ext)
    query = "?embed=1"
    return httpurl(httphp, path, query)


def printable(s):
    if type(s) != type(""):
        return s.decode("utf-8", errors="replace") if s else ""
    else:
        return s


# Transform a string so that it can be used as a file name, using
# minimserver rules: just remove any of: " * / : < > ? \ |
def tag2fn(s):
    return (
        s.replace('"', "")
        .replace("*", "")
        .replace("/", "")
        .replace(":", "")
        .replace("<", "")
        .replace(">", "")
        .replace("?", "")
        .replace("\\", "")
        .replace("|", "")
    )


def _keyvalornull(a, k):
    return a[k] if k in a else "NULL"


def _logentry(nm, e1):
    tp = _keyvalornull(e1, "tp")
    al = _keyvalornull(e1, "upnp:album")
    dr = os.path.dirname(_keyvalornull(e1, "uri"))
    tn = _keyvalornull(e1, "upnp:originalTrackNumber")
    uplog("%s tp %s alb %s dir %s tno %s" % (nm, tp, al, dr, tn))


# General container sort items comparison method
def _cmpentries_func(e1, e2):
    # uplog("cmpentries");_logentry("e1", e1);_logentry("e2", e2)
    tp1 = e1["tp"]
    tp2 = e2["tp"]
    isct1 = tp1 == "ct"
    isct2 = tp2 == "ct"

    # Containers come before items, and are sorted in alphabetic order
    ret = -2
    if isct1 and not isct2:
        ret = -1
    elif not isct1 and isct2:
        ret = 1
    elif isct1 and isct2:
        tt1 = e1["tt"]
        tt2 = e2["tt"]
        if tt1.lower() < tt2.lower():
            ret = -1
        elif tt1.lower() > tt2.lower():
            ret = 1
        else:
            ret = 0
    if ret != -2:
        # uplog("cmpentries tp1 %s tp2 %s, returning %d"%(tp1,tp2,ret))
        return ret

    # Tracks. Sort by album then directory then track number then file name
    k = "upnp:album"
    a1 = e1[k] if k in e1 else ""
    a2 = e2[k] if k in e2 else ""
    if a1 < a2:
        return -1
    elif a1 > a2:
        return 1

    a1 = os.path.dirname(e1["uri"])
    a2 = os.path.dirname(e2["uri"])
    if a1 < a2:
        return -1
    elif a1 > a2:
        return 1

    k = "upnp:originalTrackNumber"
    try:
        a1 = int(e1[k])
    except:
        a1 = 0
    try:
        a2 = int(e2[k])
    except:
        a2 = 0
    if a1 < a2:
        return -1
    elif a1 > a2:
        return 1

    # Finally: compare file names
    a1 = os.path.basename(e1["uri"])
    a2 = os.path.basename(e2["uri"])
    if a1 < a2:
        return -1
    elif a1 > a2:
        return 1

    return 0


cmpentries = functools.cmp_to_key(_cmpentries_func)


# Special comparison method for items lists: we don't want to sort by album but by title instead
def _cmpitems_func(e1, e2):

    # Tracks. Sort by title then album
    k = "tt"
    a1 = e1[k] if k in e1 else ""
    a2 = e2[k] if k in e2 else ""
    if a1 < a2:
        return -1
    elif a1 > a2:
        return 1

    k = "upnp:album"
    a1 = e1[k] if k in e1 else ""
    a2 = e2[k] if k in e2 else ""
    if a1 < a2:
        return -1
    elif a1 > a2:
        return 1

    return 0


cmpitems = functools.cmp_to_key(_cmpitems_func)


# Open embedded image. Returns mtype, size, f
def embedded_open(path):
    try:
        mutf = mutagen.File(path)
    except Exception as err:
        raise err

    f = None
    size = 0

    # First pretend that this is an ID3. These can come inside multiple file formats, don't try to
    # select on MIME, just try it.
    for tagname in mutf.keys():
        if tagname.startswith("APIC:"):
            # self.em.rclog("mp3 img: %s" % mutf[tagname].mime)
            mtype = mutf[tagname].mime
            s = mutf[tagname].data
            size = len(s)
            f = io.BytesIO(s)

    if not f:
        if "audio/flac" in mutf.mime:
            if mutf.pictures:
                mtype = mutf.pictures[0].mime
                size = len(mutf.pictures[0].data)
                f = io.BytesIO(mutf.pictures[0].data)
        elif "audio/mp4" in mutf.mime:
            if "covr" in mutf.keys():
                format = mutf["covr"][0].imageformat
                if format == mutagen.mp4.AtomDataType.JPEG:
                    mtype = "image/jpeg"
                else:
                    mtype = "image/png"
                size = len(mutf["covr"][0])
                f = io.BytesIO(mutf["covr"][0])

    if f is None:
        raise Exception(f"can't open embedded image for {path}")
    else:
        return mtype, size, f


class M3u(object):
    urlRE = re.compile(r"[a-zA-Z]+://")

    def __init__(self, fn):
        data = open(fn, "rb").read()
        try:
            data = data.decode(_g_fse, errors="surrogatescape")
        except:
            pass
        self.urls = []
        dn = os.path.dirname(os.path.abspath(fn))
        for line in data.split("\n"):
            line = line.strip(" \r")
            if not line or line[0] == "#"[0]:
                continue
            if self.urlRE.match(line):
                self.urls.append(line)
            else:
                if os.path.isabs(line):
                    self.urls.append(os.path.normpath(line))
                else:
                    self.urls.append(os.path.normpath(os.path.join(dn, line)))
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self.urls):
            raise StopIteration
        else:
            self.index += 1
            return self.urls[self.index - 1]

        
# Create bogus doc for external (http) url. This is for playlists.
def docforurl(rcldb, url):
    doc = rcldb.doc()
    doc.url = url
    doc.title = os.path.split(url)[1]
    doc.mtype = "audio/mpeg"
    return doc

        



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum


class _AudioCodecData:

    def __init__(
            self,
            codec_name: str,
            lossless: bool,
            suffixes: list[str] = None):
        self.__codec_name: str = codec_name
        self.__lossless: bool = lossless
        self.__suffixes: list[str] = (list(map(lambda x: x.lower(), suffixes))
                                      if len(suffixes if suffixes else []) > 0 else codec_name.lower())

    @property
    def codec_name(self) -> str:
        return self.__codec_name

    @property
    def lossless(self) -> str:
        return self.__lossless

    @property
    def suffixes(self) -> list[str]:
        return self.__suffixes


class AudioCodec(Enum):

    FLAC = _AudioCodecData(
        codec_name="flac",
        lossless=True,
        suffixes=["flac", "flc", "fla"])
    ALAC = _AudioCodecData(
        codec_name="alac",
        lossless=True)
    APE = _AudioCodecData(
        codec_name="ape",
        lossless=True,
        suffixes=["ape"])
    DSF = _AudioCodecData(
        codec_name="dsf",
        lossless=True)
    DFF = _AudioCodecData(
        codec_name="dff",
        lossless=True)
    WAV = _AudioCodecData(
        codec_name="wav",
        lossless=True)
    AIFF = _AudioCodecData(
        codec_name="aiff",
        lossless=True,
        suffixes=["aiff", "aif"])
    M4A = _AudioCodecData(
        codec_name="m4a",
        lossless=False)
    MP3 = _AudioCodecData(
        codec_name="mp3",
        lossless=False)
    MP2 = _AudioCodecData(
        codec_name="mp2",
        lossless=False)
    OGG = _AudioCodecData(
        codec_name="ogg",
        lossless=False)
    OPUS = _AudioCodecData(
        codec_name="opus",
        lossless=False)

    @property
    def codec_name(self) -> str:
        return self.value.codec_name

    @property
    def lossless(self) -> str:
        return self.value.lossless

    @property
    def suffixes(self) -> list[str]:
        return self.value.suffixes[:]


class LosslessStatus(Enum):

    LOSSLESS = "lossless"
    LOSSY = "lossy"
    MIXED = "mixed"


def get_lossless_status_by_value(v: str) -> LosslessStatus:
    if not v:
        raise Exception("Provide a valid lossless_status_value")
    curr: LosslessStatus
    for curr in LosslessStatus:
        if curr.value == v:
            return curr
    raise Exception(f"get_lossless_status_by_value no match for [{v}]")


def is_lossless(suffix: str):
    if not suffix:
        raise Exception("is_lossless requires a valid suffix")
    lower_suffix: str = suffix.lower()
    audio_codec: AudioCodec
    for audio_codec in AudioCodec:
        if not audio_codec.lossless:
            continue
        if lower_suffix in audio_codec.suffixes:
            return True
    return False


def is_lossy(suffix: str):
    return not is_lossless(suffix)


def get_lossless_status(suffix: str) -> LosslessStatus:
    v: bool = is_lossless(suffix)
    return LosslessStatus.LOSSLESS if (v is True) else LosslessStatus.LOSSY



# Copyright (C) 2017-2019 J.F.Dockes
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation; either version 2.1 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with this program; if not, write to the
#   Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


# This is imported from rclaudio when the audiotagfixerscript variable
# is set in the recoll configuration (uprcl sets it). It then operates
# transformations on audio tag sets, as prescribed in the minimserver
# configuration, which is pointed to by the uprclminimconfig upmpdcli
# configuration variable.

import sys
import os

from upmplgutils import getOptionValue
import minimconfig


def uplog(s):
    if not type(s) == type(b""):
        s = ("%s: %s" % ("uprcl:minimtagfixer", s)).encode("utf-8")
    sys.stderr.buffer.write(s + b"\n")
    sys.stderr.flush()


_logfp = None


def _logchange(s):
    if _logfp:
        print("%s" % s, file=_logfp)
        _logfp.flush()
    else:
        uplog(s)


class TagUpdateData(object):
    def __init__(self):
        self.tgupfile = None
        self.tgupfile = None
        self.aliastags = None

        self.minimcnffn = getOptionValue("uprclminimconfig")
        if not self.minimcnffn:
            return

        self.minimconfig = minimconfig.MinimConfig(self.minimcnffn)
        self.tgupfile = self.minimconfig.getsimplevalue("minimserver.tagUpdate")
        # uplog("Minim config read: tagUpdate: %s" % self.tgupfile)
        if self.tgupfile:
            self.tgupfile = self._makeabs(self.tgupfile)
            if not os.path.exists(self.tgupfile):
                uplog(f"gettagupdate: {self.tgupfile} does not exist")
                return
        wlogfn = self.minimconfig.getsimplevalue("minimserver.writeTagChanges")
        if wlogfn:
            wlogfn = self._makeabs(wlogfn)
            global _logfp
            try:
                _logfp = open(wlogfn, "a")
            except Exception as ex:
                uplog(f"minimtagfixer: can't open {wlogfn} : {ex}")
        self.aliastags = self.minimconfig.getaliastags()
        self.tagvalue = self.minimconfig.gettagvalue()

    # The defs and log files are to be found in the same dir as the
    # logs (see minim doc). So compute minimserver/etc/xx.conf/../../data/fn
    # This probably only works for linux
    def _makeabs(self, fn):
        if not fn:
            return fn
        if not os.path.isabs(fn):
            dir = os.path.dirname(os.path.dirname(self.minimcnffn))
            fn = os.path.join(os.path.join(dir, "data"), fn)
        return fn

    # https://minimserver.com/ug-other.html#Tag%20update
    # Tag value replacement specs, e.g. "Johann Sebastian Bach" -> "Bach J.S."
    def gettagupdate(self):
        if not self.tgupfile:
            return ()
        try:
            f = open(self.tgupfile, "rb")
        except Exception as ex:
            uplog(f"gettagupdate: can't open {self.tgupfile} for reading: {ex}")
            return

        groups = []
        seltags = []
        filtertags = []
        tagadds = []
        tagdels = []
        seltagsdone = False
        for line in f.readlines():
            line = line.strip().decode("utf-8")
            if not line or line[0] not in ("@"[0], "&"[0], "="[0], "-"[0], "+"[0]):
                # uplog("gettaupdate: skipping [%s]" % line)
                continue

            if line[0] != "@"[0]:
                seltagsdone = True

            # Regularise:
            if line[0] == "-" and line[-1] != "="[0]:
                line += "="
            # Compute tag and value
            eq = line[1:].find("=")
            tagname = line[1 : eq + 1].strip().lower()
            tagval = line[eq + 2 :].strip().encode("utf-8")
            # uplog("gettaupdate: tagname [%s] tagval [%s]"%(tagname,tagval))

            if line[0] == "@"[0]:
                if seltagsdone:
                    # We've already seen @tag(s) and at least one
                    # other directive, this group is done, finalize it
                    # before beginning the next
                    groups.append((seltags, filtertags, tagadds, tagdels))
                    seltags = []
                    filtertags = []
                    tagadds = []
                    tagdels = []
                    seltagsdone = False
                seltags.append((tagname, tagval))
            elif line[0] == "&"[0]:
                filtertags.append((tagname, tagval))
            elif line[0] == "="[0]:
                tagadds.append((tagname, tagval))
            elif line[0] == "-"[0]:
                tagdels.append((tagname, ""))
            elif line[0] == "+"[0]:
                tagadds.append((tagname, tagval))

        if seltagsdone:
            groups.append((seltags, filtertags, tagadds, tagdels))

        return groups


tud = TagUpdateData()
groups = tud.gettagupdate()
# uplog("minimtagfixer: groups %s" % groups)


def tagupdate(tags):
    for group in groups:
        # Must match at least one element of group[0]
        sel = False
        for tag, val in group[0]:
            if tag in tags and tags[tag] == val:
                sel = True
                break
        if not sel:
            return
        # Must match all in group[1]
        for tag, val in group[1]:
            if not tag in tags or tags[tag] != val:
                sel = False
        if not sel:
            return
        # Apply adds/mods
        for tag, val in group[2]:
            old = tags[tag] if tag in tags else ""
            _logchange("Setting [%s] from [%s] to [%s]" % (tag, old, val))
            tags[tag] = val
        # Apply dels
        for tag, val in group[3]:
            _logchange("Clearing [%s]" % tag)
            del tags[tag]


def tagvalue(tags):
    pass


def aliastags(tags):
    # uplog(f"ALIASTAGS: {tud.aliastags}")
    if tud.aliastags:
        for orig, target, rep in tud.aliastags:
            try:
                val = tags[orig]
                targetexists = "yes" if target in tags else "no"
                # uplog(f"aliastags: Rep {rep} tags[{orig}]=[{val}] {target} in tags: {targetexists}")
                if val and (rep or not target in tags or not tags[target]):
                    # uplog(f"tags[{target}] -> {val}")
                    tags[target] = val
            except Exception as ex:
                # uplog(f"EXCEPTION: {ex}")
                pass


def tagfix(tags={}):
    # uplog("tagfix(%s)"%tags)
    tagvalue(tags)
    aliastags(tags)
    tagupdate(tags)



# Copyright (C) 2021-2022 J.F.Dockes
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the
#   Free Software Foundation, Inc.,
#   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import os
import subprocess
import threading

from upmplgutils import uplog, direntry, getOptionValue
import conftree


class UpmpdcliRadios(object):
    def __init__(self, upconfig):
        datadir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        if not os.path.isabs(datadir):
            datadir = "/usr/share/upmpdcli"
        self.fetchstream = os.path.join(datadir, "rdpl2stream", "fetchStream.py")
        self._radios = []
        self._readRadios(upconfig)
        # uplog("Radios: %s" % self._radios)

    def _fetchStream(self, index, title, uri, artUri, mime):
        uplog(f"upradios: Fetching {title}")
        streamUri = ""
        try:
            streamUri = subprocess.check_output([self.fetchstream, uri])
            streamUri = streamUri.decode("utf-8").strip("\r\n")
        except Exception as ex:
            uplog("fetchStream.py failed for %s: %s" % (title, ex))
        if streamUri:
            self._radios[index] = (title, streamUri, uri, artUri, mime)

    def _readRadiosFromConf(self, conf):
        """Read radio definitions from a config file (either main file or radiolist)"""
        keys = conf.getSubKeys_unsorted()
        threads = []
        cntt = 0
        for k in keys:
            if k.startswith("radio"):
                title = k[6:]
                uri = conf.get("url", k)
                artUri = conf.get("artUrl", k)
                mime = conf.get("mime", k)
                self._radios.append(None)
                idx = len(self._radios) - 1
                t = threading.Thread(
                    target=self._fetchStream, args=(idx, title, uri, artUri, mime)
                )
                t.start()
                threads.append(t)
                cntt += 1
                # Note that it's ok to join a thread multiple times, so we keep things simple
                if cntt >= self._maxthreads:
                    uplog(f"upradios: Waiting for threads")
                    for t in threads:
                        t.join()
                    uplog(f"upradios: Waiting done")
                    cntt = 0
        uplog(f"upradios: Waiting for threads")
        for t in threads:
            t.join()
        uplog(f"upradios: Waiting done")
        # Get rid of None entries in the list, keeping the order
        tlist = []
        for rd in self._radios:
            if rd:
                tlist.append(rd)
        self._radios = tlist
        uplog(f"upradios: Init done")

    def _readRadios(self, upconfig):
        """Read radio definitions from main config file, then possible radiolist"""
        self._radios = []
        self._maxthreads = upconfig.get("upradiosmaxthreads")
        if self._maxthreads:
            self._maxthreads = int(self._maxthreads)
        else:
            self._maxthreads = 5
        self._readRadiosFromConf(upconfig)
        radiolist = getOptionValue(
            "radiolist", "/usr/share/upmpdcli/radio_scripts/radiolist.conf"
        )
        if radiolist:
            radioconf = conftree.ConfSimple(radiolist)
            self._readRadiosFromConf(radioconf)

    def __iter__(self):
        return RadioIterator(self)

    # Note: not implementing __getitem__ because don't want to deal with slices, neg indexes etc.
    def get_radio(self, idx):
        if idx < 0 or idx > len(self._radios):
            raise (IndexError(f"bad radio index {idx}"))
        radio = self._radios[idx]
        return {
            "index": idx,
            "title": radio[0],
            "streamUri": radio[1],
            "uri": radio[2],
            "artUri": radio[3],
            "mime": radio[4],
        }


class RadioIterator:
    def __init__(self, radios):
        self._radios = radios
        self._index = 0

    def __next__(self):
        if self._index < (len(self._radios._radios)):
            result = self._radios.get_radio(self._index)
            self._index += 1
            return result

        raise StopIteration


def radioToEntry(pid, id, radio):
    # uplog(f"radioToEntry: pid {pid} id {id}")
    # if this comes from a 'browse meta', the id is set, else compute it
    if id is None:
        objid = pid
        if objid[-1] != "$":
            objid += "$"
        objid += "e" + str(radio["index"])
    else:
        objid = id
    if "mime" in radio and radio["mime"]:
        mime = radio["mime"]
    else:
        mime = "audio/mpeg"
    return {
        "pid": pid,
        "id": objid,
        "uri": radio["streamUri"],
        "tp": "it",
        "res:mime": mime,
        "upnp:class": "object.item.audioItem.musicTrack",
        "upnp:albumArtURI": radio["artUri"],
        "tt": radio["title"],
        # This is for Kodi mostly, to avoid displaying a big "Unknown" placeholder
        "upnp:artist": "Internet Radio",
        "upnp:album": "upmpdcli",
    }


def radioIndexFromId(objid):
    stridx = objid.find("$e")
    if stridx == -1:
        raise Exception(f"Bad objid {objid}")
    return int(objid[stridx + 2 :])


if __name__ == "__main__":
    from upmplgutils import getConfigObject

    radios = UpmpdcliRadios(getConfigObject())
    for radio in radios:
        print("%s" % radio)



# Copyright (C) 2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class SongInfo:

    def __init__(
            self,
            song_id: str,
            bitrate: int,
            bit_depth: int,
            sampling_rate: int,
            suffix: str,
            duration: int):
        self.__song_id: str = song_id
        self.__bit_depth: int = bit_depth
        self.__bitrate: int = bitrate
        self.__sampling_rate: int = sampling_rate
        self.__suffix: str = suffix
        self.__duration: int = duration

    @property
    def song_id(self) -> str:
        return self.__song_id

    @property
    def bitrate(self) -> int:
        return self.__bitrate

    @property
    def bit_depth(self) -> int:
        return self.__bit_depth

    @property
    def sampling_rate(self) -> int:
        return self.__sampling_rate

    @property
    def suffix(self) -> str:
        return self.__suffix

    @property
    def duration(self) -> int:
        return self.__duration



##########################################################################
# Copyright 2009 Carlos Ribeiro
#
# This file is part of Radio Tray
#
# Radio Tray is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 1 of the License, or
# (at your option) any later version.
#
# Radio Tray is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radio Tray.  If not, see <http://www.gnu.org/licenses/>.
#
##########################################################################
import sys
from urllib.request import Request as UrlRequest
from urllib.request import urlopen as urlUrlopen
from common import USER_AGENT
import logging

class AsfPlaylistDecoder:
    def __init__(self):
        self.log = logging.getLogger('upmpdcli')

    def isStreamValid(self, contentType, firstBytes):
        if 'video/x-ms-asf' in contentType and \
           firstBytes.strip().lower().startswith('b[reference]'):
            self.log.info('Stream is readable by ASF Playlist Decoder')
            return True
        else:
            return False

        
    def extractPlaylist(self,  url):
        self.log.info('ASF: downloading playlist..')
        req = UrlRequest(url)
        req.add_header('User-Agent', USER_AGENT)
        f = urlUrlopen(req)
        str = f.read()
        f.close()

        self.log.info('ASF: playlist downloaded, decoding...')

        playlist = []
        lines = str.splitlines()
        for line in lines:
            if line.startswith(b"Ref"):
                list = line.split(b"=", 1)
                tmp = list[1].strip()
                if tmp.endswith(b"?MSWMExt=.asf"):
                    playlist.append(tmp.replace(b"http", b"mms"))
                else:
                    playlist.append(tmp)
         
        return playlist



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class ArtistIdName:

    def __init__(
            self,
            artist_id: str,
            artist_name: str):
        self.__artist_id: str = artist_id
        self.__artist_name: str = artist_name

    @property
    def artist_id(self) -> str:
        return self.__artist_id

    @property
    def artist_name(self) -> str:
        return self.__artist_name


class ArtistIdNameCoverArt(ArtistIdName):

    def __init__(
            self,
            artist_id: str,
            artist_name: str,
            cover_art: str = None):
        super().__init__(
            artist_id=artist_id,
            artist_name=artist_name)
        self.__cover_art: str = cover_art

    @property
    def cover_art(self) -> str:
        return self.__cover_art



#!/usr/bin/env python3
from __future__ import print_function

# Metadata getter for Radio Paradiase stations
# Script based on meta-fip.py script (https://www.lesbonscomptes.com/upmpdcli/)
#
# 'channelid' is first parameter (set in config file)

import json
import requests
import sys
import time

def elemNumber(objN, elemN):
  if elemN in objN:
      r = int(objN[elemN])
      if r >= 0:
        return r
  return 0

def elemText(objT, elemT):
  if elemT in objT:
    return str(objT[elemT])
  return None

def return_metadata(title, artist, album, year, artUrl, reload):
  metadata = {'title'  : display_title,
              'artist' : artist,
              'album'  : album,
              'year'   : year,
              'artUrl' : artUrl,
              'reload' : reload}
  print("%s"% json.dumps(metadata))
  return

def return_no_metadata(reload):
  metadata = {'title'  : None,
              'artist' : None,
              'album'  : None,
              'year'   : None,
              'artUrl' : None,
              'reload' : reload}
  print("%s"% json.dumps(metadata))
  return

def disp_title(artist, title, year, album):
  ab = ''
  yr = ''
  dt = ''

  if year:
    yr += ' (' + str(year) + ')'

  if album:
    ab += ' [' + album + ']'

  if   title and artist and year:
    dt = artist + ' – ' + title + yr + ab
  elif title and artist:
    dt = artist + ' – ' + title + ab
  elif title and year:
    dt = title + yr + ab
  elif title:
    dt = title + ab
  elif artist and album:
    dt = artist + + ' – Album: ' + ab + yr
  elif artist:
    dt = artist + yr
  elif album:
    dt = 'Album: ' + ab + yr
  return dt



if len(sys.argv) > 1:
  channelid = int(sys.argv[1])
else:
  sys.exit(1)

try:
  r = requests.get('https://api.radioparadise.com/api/now_playing?chan=' +
                   str(channelid))
  r.raise_for_status()
except:
  return_no_metadata(10)

try:
  song = r.json()
except:
  return_no_metadata(120)


now  = int(time.time())
time = elemNumber(song, 'time')
#print('now %d time %d' % (now, time), file=sys.stderr)
if time > 2 :
  title   = elemText(song, 'title')
  artist  = elemText(song, 'artist')
  album   = elemText(song, 'album')
  year    = elemText(song, 'year')
  artUrl  = elemText(song, 'cover')

  # For radio streams title is displayed twice (in both artist
  # and title fields),
  # in order to display artist as well, join 'title' and 'artist'
  display_title = disp_title(artist, title, year, album)
  return_metadata(display_title, artist, album, year, artUrl, time + 1)
else:
  return_no_metadata(time + 1)



# Copyright (C) 2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class _CacheTypeData:

    def __init__(self, cache_name: str):
        self.__cache_name: str = cache_name

    @property
    def cache_name(self) -> str:
        return self.__cache_name


class CacheType(Enum):

    GENERIC_KEY_VALUE = _CacheTypeData("generic_kv")
    ITEM_IDENTIFIER_CODEC = _CacheTypeData("iid_codec")

    @property
    def cache_name(self) -> str:
        return self.value.cache_name


# duplicate check
name_checker_set: set[str] = set()
id_checker_set: set[int] = set()
for v in CacheType:
    if v.cache_name in name_checker_set:
        raise Exception(f"Duplicated name [{v.cache_name}]")
    name_checker_set.add(v.cache_name)
    id_checker_set.add(v.cache_name)



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum


class _AlbumPropertyKeyData:

    def __init__(
            self,
            property_key: str = None,
            display_value: str = None,
            unique_value: bool = False,
            max_items: int = 100):
        self.__property_key: str = property_key
        self.__display_value: str = display_value
        self.__unique_value: bool = unique_value
        self.__max_items: int = max_items

    @property
    def property_key(self) -> str:
        return self.__property_key

    @property
    def display_value(self) -> str:
        return self.__display_value

    @property
    def unique_value(self) -> bool:
        return self.__unique_value

    @property
    def max_items(self) -> int:
        return self.__max_items


class AlbumPropertyKey(Enum):

    GENRE = _AlbumPropertyKeyData(display_value="Genre")
    MOOD = _AlbumPropertyKeyData(display_value="Mood")
    LABEL_INITIAL = _AlbumPropertyKeyData(display_value="Label (Initial)")
    LABEL = _AlbumPropertyKeyData(display_value="Label")
    DECADE = _AlbumPropertyKeyData(display_value="Decade", unique_value=True)
    YEAR = _AlbumPropertyKeyData(display_value="Year", unique_value=True)
    ALBUM_ARTIST = _AlbumPropertyKeyData(display_value="Album Artist")
    ARTIST = _AlbumPropertyKeyData(display_value="Artist")
    ARTIST_CONTRIBUTOR = _AlbumPropertyKeyData(display_value="Artist/Contributor")
    CONTRIBUTOR = _AlbumPropertyKeyData(display_value="Contributor")
    RELEASE_TYPE = _AlbumPropertyKeyData(display_value="Release Type")
    SUFFIX = _AlbumPropertyKeyData(display_value="Suffix")
    LOSSLESS_STATUS = _AlbumPropertyKeyData(display_value="Compression Type", unique_value=True)
    QUALITY_BADGE = _AlbumPropertyKeyData(display_value="Quality Badge", unique_value=True)
    BIT_DEPTH = _AlbumPropertyKeyData(display_value="Bit Depth")
    SAMPLING_RATE = _AlbumPropertyKeyData(display_value="Sampling Rate")
    CHANNEL_COUNT = _AlbumPropertyKeyData(display_value="Channel Count")
    HAS_MUSICBRAINZ = _AlbumPropertyKeyData(display_value="MusicBrainz Album Id")
    HAS_COVER_ART = _AlbumPropertyKeyData(display_value="Cover Art")

    @property
    def property_key(self) -> str:
        return self.value.property_key if self.value.property_key else self.name.lower()

    @property
    def display_value(self) -> str:
        return self.value.display_value if self.value.display_value else self.name

    @property
    def unique_value(self) -> bool:
        return self.value.unique_value

    @property
    def max_items(self) -> int:
        return self.value.max_items


def get_album_property_key(property_key: str) -> AlbumPropertyKey:
    apk: AlbumPropertyKey
    for apk in AlbumPropertyKey:
        if property_key == apk.property_key:
            return apk
    # no match
    return None


class AlbumPropertyKeyValue:

    def __init__(
            self,
            key: str,
            value: str):
        self.__key: str = key
        self.__value: str = value

    @property
    def key(self) -> str:
        return self.__key

    @property
    def value(self) -> str:
        return self.__value


class AlbumPropertyKeyOccurrence:

    def __init__(
            self,
            property_key: str,
            property_value_count: int,
            is_missing_for_some: bool,
            representative_album_id: str):
        self.__property_key: str = property_key
        self.__property_value_count: int = property_value_count
        self.__is_missing_for_some: bool = is_missing_for_some
        self.__representative_album_id: str = representative_album_id

    @property
    def property_key(self) -> str:
        return self.__property_key

    @property
    def property_value_count(self) -> int:
        return self.__property_value_count

    @property
    def is_missing_for_some(self) -> bool:
        return self.__is_missing_for_some

    @property
    def representative_album_id(self) -> str:
        return self.__representative_album_id


def condition_list_contains_negative(condition_list: list[AlbumPropertyKeyValue], album_property_key: str) -> bool:
    curr: AlbumPropertyKeyValue
    for curr in condition_list if condition_list else []:
        if curr.key == album_property_key and curr.value is None:
            return True
    return False


def condition_list_positive_count(condition_list: list[AlbumPropertyKeyValue], album_property_key: str) -> int:
    cnt: int = 0
    curr: AlbumPropertyKeyValue
    for curr in condition_list if condition_list else []:
        if curr.key == album_property_key and curr.value is not None:
            cnt += 1
    return cnt


def condition_list_contains_positive(
        condition_list: list[AlbumPropertyKeyValue],
        album_property_key: str,
        album_property_value: str) -> bool:
    curr: AlbumPropertyKeyValue
    for curr in condition_list if condition_list else []:
        if curr.key == album_property_key and curr.value == album_property_value:
            return True
    return False


class AlbumPropertyValueOccurrence:

    def __init__(
            self,
            property_value: str,
            album_count: int,
            is_missing_for_some: bool,
            representative_album_id: str):
        self.__property_value: str = property_value
        self.__album_count: int = album_count
        self.__is_missing_for_some: bool = is_missing_for_some
        self.__representative_album_id: str = representative_album_id

    @property
    def property_value(self) -> str:
        return self.__property_value

    @property
    def album_count(self) -> int:
        return self.__album_count

    @property
    def is_missing_for_some(self) -> bool:
        return self.__is_missing_for_some

    @property
    def representative_album_id(self) -> str:
        return self.__representative_album_id



# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum

class TagType(Enum):
    
    ALL_STREAMS = 100, "all", "All Streams"
    BY_TITLE = 101, "byt", "By Title"
    BY_CODEC = 102, "byc", "By Codec"

    def __init__(self, 
            num: int, 
            tag_name: str, 
            tag_title: str):
        self.num: int = num
        self.tag_name: str = tag_name
        self.tag_title: str = tag_title

    def getTagName(self) -> str:
        return self.tag_name

    def getTagTitle(self) -> str:
        return self.tag_title

def get_tag_type_by_name(tag_name: str) -> TagType:
    #msgproc.log(f"get_tag_Type_by_name with {tag_name}")
    for _, member in TagType.__members__.items():
        if tag_name == member.getTagName():
            return member
    raise Exception(f"get_tag_type_by_name with {tag_name} NOT found")

# duplicate check
name_checker_set: set[str] = set()
id_checker_set: set[int] = set()
for v in TagType:
    if v.getTagName() in name_checker_set:
        raise Exception(f"Duplicated name [{v.getTagName()}]")
    if v.value[0] in id_checker_set:
        raise Exception(f"Duplicated id [{v.value[0]}]")
    name_checker_set.add(v.getTagName())
    id_checker_set.add(v.value[0])


# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum

class ItemIdentifierKey(Enum):
    
    THING_NAME = 0, 'n'
    THING_VALUE = 1, 'v'
            
    def __init__(self, 
            num: int, 
            key_name: str):
        self.num: int = num
        self.key_name: str = key_name
    
    def getName(self) -> str:
        return self.key_name
    
# duplicate check
name_checker_set: set[str] = set()
id_checker_set: set[int] = set()
for v in ItemIdentifierKey:
    if v.getName() in name_checker_set:
        raise Exception(f"Duplicated name [{v.getName()}]")
    if v.value[0] in id_checker_set:
        raise Exception(f"Duplicated id [{v.value[0]}]")
    name_checker_set.add(v.getName())
    id_checker_set.add(v.value[0])


# Copyright (C) 2024 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class ContextKey(Enum):
    CANNOT_GET_STREAM_INFO = 0, "cannot-get-stream-info", False
    SUCCESS_COUNT = 1, "success-count", 0
    PROCESS_COUNT = 2, "process-count", 0
    PLAYED_ALBUM_TRACKS_DICT = 3, "played-albums-tracks-dict", dict()
    KNOWN_TRACKS_COUNT = 4, "known-tracks-count", 0
    GUESSED_TRACKS_COUNT = 5, "guessed-tracks-count", 0
    IS_ALBUM = 6, "is-album", False
    ALBUM_FIRST_TRACK_BIT_DEPTH = 7, "album-first-track-bit-depth", None
    ALBUM_FIRST_TRACK_SAMPLE_RATE = 8, "album-first-track-sample-rate", None
    ALBUM_FIRST_TRACK_AUDIO_QUALITY = 9, "album-first-track-audio-quality", None
    ASSUMED_FROM_FIRST_ALBUM_TRACK_COUNT = 10, "assumed_from_first_album_track_count", 0
    GET_STREAM_COUNT = 11, "get-stream-count", 0
    IS_PLAYLIST = 12, "is-playlist", False
    IS_MIX = 13, "is-mix", False
    IS_TRACK = 19, "is-track", False
    ASSUMED_BY_MAX_AUDIO_QUALITY_COUNT = 14, "assumed-by-max-audio-quality-count", 0
    KNOWN_TRACK_DICT = 15, "known-track-dict", None
    ASSUMED_FROM_FIRST_ALBUM_TRACK_DICT = 16, "assumed-from-first-album-track-dict", None
    GET_STREAM_DICT = 17, "get-stream-dict", None
    GUESSED_TRACK_DICT = 18, "guessed-track-dict", None

    def __init__(self,
            num: int,
            key_name: str,
            default_value: any):
        self.__num: int = num
        self.__key_name: str = key_name
        self.__default_value: any = default_value

    @property
    def name(self) -> str:
        return self.__key_name

    @property
    def default_value(self) -> any:
        return self.__default_value


# duplicate check
name_checker_set: set[str] = set()
id_checker_set: set[int] = set()
for v in ContextKey:
    if v.name in name_checker_set:
        raise Exception(f"Duplicated name [{v.name}]")
    if v.value[0] in id_checker_set:
        raise Exception(f"Duplicated id [{v.value[0]}]")
    name_checker_set.add(v.name)
    id_checker_set.add(v.value[0])



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from metadata_model import AlbumPropertyMetaModel
from typing import Any
import datetime


class AlbumPropertyMetadata:

    def __init__(
            self,
            album_id: str,
            key: str,
            value: str,
            created: datetime.datetime,
            updated: datetime.datetime):
        self.__album_id: str = album_id
        self.__key: str = key
        self.__value: str = value
        self.__created: datetime.datetime = created
        self.__updated: datetime.datetime = updated

    def set_value(self, album_metadata_model: AlbumPropertyMetaModel, value: Any):
        return self._set(album_metadata_model.column_name, value)

    @property
    def album_id(self) -> str:
        return self.__album_id

    @property
    def album_property_key(self) -> str:
        return self.__key

    @property
    def album_property_value(self) -> str:
        return self.__value

    @property
    def created_timestamp(self) -> datetime.datetime:
        return self.__created

    @property
    def updated_timestamp(self) -> datetime.datetime:
        return self.__updated



##########################################################################
# Copyright 2009 Carlos Ribeiro
#
# This file is part of Radio Tray
#
# Radio Tray is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 1 of the License, or
# (at your option) any later version.
#
# Radio Tray is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radio Tray.  If not, see <http://www.gnu.org/licenses/>.
#
##########################################################################
import sys
from urllib.request import Request as UrlRequest
from urllib.request import urlopen as urlUrlopen
import xml.etree.ElementTree as ET
from io import BytesIO
import logging
from common import USER_AGENT

class XspfPlaylistDecoder:
    def __init__(self):
        self.log = logging.getLogger('upmpdcli')

    def isStreamValid(self, contentType, firstBytes):
        if 'application/xspf+xml' in contentType:
            self.log.info('Stream is readable by XSPF Playlist Decoder')
            return True
        else:
            return False


    def extractPlaylist(self,  url):
        self.log.info('XSPF: downloading playlist...')
        req = UrlRequest(url)
        req.add_header('User-Agent', USER_AGENT)
        f = urlUrlopen(req)
        str = f.read()
        f.close()
        self.log.info('XSPF: playlist downloaded, decoding...')

        root = ET.parse(BytesIO(str))
        ns = {'xspf':'http://xspf.org/ns/0/'}
        elements = root.findall(".//xspf:track/xspf:location", ns)

        result = []
        for r in elements:
            result.append(r.text)

        return result



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum

class _ArtistRoleData:

    def __init__(
            self,
            key: str,
            display_value: str):
        self.__key: str = key
        self.__display_value: str = display_value
    
    @property
    def key(self) -> str:
        return self.__key

    @property
    def display_value(self) -> str:
        return self.__display_value
    

class ArtistRole(Enum):

    ALBUM_ARTIST = _ArtistRoleData(
        key="albumartist",
        display_value="Album Artist")
    
    @property
    def key(self) -> str:
        return self.value.key

    @property
    def display_value(self) -> str:
        return self.value.display_value
        

def get_artist_role_display_value(key: str) -> str:
    curr: ArtistRole
    for curr in ArtistRole:
        if curr.key == key:
            if curr.display_value is None or len(curr.display_value) == 0:
                break
            return curr.display_value
    return key.capitalize()



# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum
from radio_station_entry import RadioStationEntry


def get_base_url() -> str:
    return "stream.motherearthradio.de"


def get_base_listen_url() ->  str: 
    return f"https://{get_base_url()}/listen"


class Codec(Enum):
    FLAC = "flac"
    AAC = "aac"
    MP3 = "mp3"


class MimeType(Enum):
    FLAC = "audio/flac"
    AAC = "audio/aac"
    MP3 = "audio/mpeg"


class RadioTitle(Enum):
    MAIN = "Mother Earth Main"
    JAZZ = "Mother Earth Jazz"
    KLASSIK = "Mother Earth Klassik"
    INSTRUMENTAL = "Mother Earth Instrumental"


class RadioStation(Enum):

    MAIN_FLAC_MONO = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth/motherearth.mono",
        title=RadioTitle.MAIN.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=192000,
        channel_count=1)
    MAIN_FLAC_192_24 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth/motherearth",
        title=RadioTitle.MAIN.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=192000)
    MAIN_FLAC_96_24 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth/motherearth.flac-lo",
        title=RadioTitle.MAIN.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=96000)
    MAIN_AAC_24_96 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth/motherearth.aac",
        title=RadioTitle.MAIN.value,
        codec=Codec.AAC.value,
        mimetype=MimeType.AAC.value,
        bit_depth=24,
        sampling_rate=96000,
        channel_count=2)
    MAIN_AAC_16_48 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth/motherearth.aac-lo",
        title=RadioTitle.MAIN.value,
        codec=Codec.AAC.value,
        mimetype=MimeType.AAC.value,
        bit_depth=24,
        sampling_rate=48000,
        channel_count=2)
    MAIN_MP3_16_44 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth/motherearth.mp3",
        title=RadioTitle.MAIN.value,
        codec=Codec.MP3.value,
        mimetype=MimeType.MP3.value,
        bit_depth=16,
        sampling_rate=44100,
        channel_count=2)

    JAZZ_FLAC_MONO = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_jazz/motherearth.jazz.mono",
        title=RadioTitle.JAZZ.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=192000,
        channel_count=1)
    JAZZ_FLAC_192_24 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_jazz/motherearth.jazz",
        title=RadioTitle.JAZZ.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=192000)
    JAZZ_FLAC_96_24 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_jazz/motherearth.jazz.flac-lo",
        title=RadioTitle.JAZZ.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=96000)
    JAZZ_AAC_24_96 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_jazz/motherearth.jazz.mp4",
        title=RadioTitle.JAZZ.value,
        codec=Codec.AAC.value,
        mimetype=MimeType.AAC.value,
        bit_depth=24,
        sampling_rate=96000,
        channel_count=2)
    JAZZ_AAC_16_48 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_jazz/motherearth.jazz.aac-lo",
        title=RadioTitle.JAZZ.value,
        codec=Codec.AAC.value,
        mimetype=MimeType.AAC.value,
        bit_depth=24,
        sampling_rate=48000,
        channel_count=2)
    JAZZ_MP3_16_44 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_jazz/motherearth.jazz.mp3",
        title=RadioTitle.JAZZ.value,
        codec=Codec.MP3.value,
        mimetype=MimeType.MP3.value,
        bit_depth=16,
        sampling_rate=44100,
        channel_count=2)

    CLASSIC_FLAC_MONO = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_klassik/motherearth.klassik.mono",
        title=RadioTitle.KLASSIK.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=192000,
        channel_count=1)
    CLASSIC_FLAC_192_24 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_klassik/motherearth.klassik",
        title=RadioTitle.KLASSIK.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=192000)
    CLASSIC_FLAC_96_24 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_klassik/motherearth.klassik.flac-lo",
        title=RadioTitle.KLASSIK.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=96000)
    CLASSIC_AAC_24_96 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_klassik/motherearth.klassik.aac",
        title=RadioTitle.KLASSIK.value,
        codec=Codec.AAC.value,
        mimetype=MimeType.AAC.value,
        bit_depth=24,
        sampling_rate=96000,
        channel_count=2)
    CLASSIC_AAC_16_48 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_klassik/motherearth.klassik.aac-lo",
        title=RadioTitle.KLASSIK.value,
        codec=Codec.AAC.value,
        mimetype=MimeType.AAC.value,
        bit_depth=24,
        sampling_rate=48000,
        channel_count=2)
    CLASSIC_MP3_16_44 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_klassik/motherearth.klassik.mp3",
        title=RadioTitle.KLASSIK.value,
        codec=Codec.MP3.value,
        mimetype=MimeType.MP3.value,
        bit_depth=16,
        sampling_rate=44100,
        channel_count=2)


    INSTRUMENTAL_FLAC_MONO = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_instrumental/motherearth.instrumental.mono",
        title=RadioTitle.INSTRUMENTAL.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=192000,
        channel_count=1)
    INSTRUMENTAL_FLAC_192_24 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_instrumental/motherearth.instrumental",
        title=RadioTitle.INSTRUMENTAL.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=192000)
    INSTRUMENTAL_FLAC_96_24 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_instrumental/motherearth.instrumental.flac-lo",
        title=RadioTitle.INSTRUMENTAL.value,
        codec=Codec.FLAC.value,
        mimetype=MimeType.FLAC.value,
        bit_depth=24,
        sampling_rate=96000)
    INSTRUMENTAL_AAC_24_96 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_instrumental/motherearth.instrumental.aac",
        title=RadioTitle.INSTRUMENTAL.value,
        codec=Codec.AAC.value,
        mimetype=MimeType.AAC.value,
        bit_depth=24,
        sampling_rate=96000,
        channel_count=2)
    INSTRUMENTAL_AAC_16_48 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_instrumental/motherearth.instrumental.aac-lo",
        title=RadioTitle.INSTRUMENTAL.value,
        codec=Codec.AAC.value,
        mimetype=MimeType.AAC.value,
        bit_depth=24,
        sampling_rate=48000,
        channel_count=2)
    INSTRUMENTAL_MP3_16_44 = RadioStationEntry(
        url=f"{get_base_listen_url()}/motherearth_instrumental/motherearth.instrumental.mp3",
        title=RadioTitle.INSTRUMENTAL.value,
        codec=Codec.MP3.value,
        mimetype=MimeType.MP3.value,
        bit_depth=16,
        sampling_rate=44100,
        channel_count=2)


def get_by_name(station_name: str) -> RadioStationEntry:
    curr: RadioStationEntry
    for curr in RadioStation:
        if curr.name == station_name:
            return curr
    return None


radio_station_list: list[RadioStationEntry] = list([x.value for x in RadioStation])




# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class _TagTypeData:

    def __init__(self, name: str, title: str, prefer_non_static_icon: bool = False):
        self.__name: str = name
        self.__title: str = title
        self.__prefer_non_static_icon: bool = prefer_non_static_icon

    @property
    def name(self) -> str:
        return self.__name

    @property
    def title(self) -> str:
        return self.__title

    @property
    def prefer_non_static_icon(self) -> bool:
        return self.__prefer_non_static_icon


class TagType(Enum):
    PAGE_SELECTION = _TagTypeData("pgslctn", "Page Selection")
    CATEGORIES = _TagTypeData("ctgrs", "Categories")
    HOME = _TagTypeData("home", "Home")
    EXPLORE = _TagTypeData("xplr", "Explore")
    EXPLORE_NEW_MUSIC = _TagTypeData("new", "New Music")
    EXPLORE_TIDAL_RISING = _TagTypeData("rising", "Tidal Rising")
    FOR_YOU = _TagTypeData("foru", "For You")
    HIRES_PAGE = _TagTypeData("hires", "Hi-Res")
    GENRES_PAGE = _TagTypeData("genres", "Genres")
    LOCAL_GENRES_PAGE = _TagTypeData("local_genres", "Local Genres")
    MOODS_PAGE = _TagTypeData("moods", "Moods")
    ALL_PLAYLISTS = _TagTypeData("allplsts", "Playlists", prefer_non_static_icon=True)
    MY_PLAYLISTS = _TagTypeData("myplsts", "My Playlists", prefer_non_static_icon=True)
    FAVORITE_ARTISTS = _TagTypeData("fvrtsts", "My Artists", prefer_non_static_icon=True)
    FAVORITE_ALBUMS = _TagTypeData("fvlbms", "My Albums", prefer_non_static_icon=True)
    FAVORITE_TRACKS = _TagTypeData("fvrttrks", "My Tracks", prefer_non_static_icon=True)
    BOOKMARKS = _TagTypeData("lstnq", "Bookmarks", prefer_non_static_icon=True)
    PLAYBACK_STATISTICS = _TagTypeData("plbkstts", "Playback Statistics", prefer_non_static_icon=True)

    @property
    def name(self) -> str:
        return self.value.name

    @property
    def title(self) -> str:
        return self.value.title

    @property
    def prefer_non_static_icon(self) -> bool:
        return self.value.prefer_non_static_icon


def get_tidal_tag_type_by_name(tag_name: str) -> TagType:
    for _, member in TagType.__members__.items():
        if tag_name == member.name:
            return member
    raise Exception(f"get_tidal_tag_type_by_name with {tag_name} NOT found")



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
from enum import Enum
from functools import cmp_to_key
import copy
import time

from subsonic_connector.song import Song
from subsonic_connector.album import Album

from album_metadata import AlbumMetadata
from metadata_model import AlbumMetadataModel

from codec_delimiter_style import CodecDelimiterStyle
import constants
import config
import persistence_constants
from release_date import ReleaseDate
from msgproc_provider import msgproc


__split_characters: list[str] = [' ', '-', '_']


def split_string(string, delimiters):
    pattern = r'|'.join(delimiters)
    return re.split(pattern, string)


class Starter(Enum):
    CD = 0
    DISC = 1
    DISCO = 2
    DISK = 3
    D = 4
    VOLUME = 5
    VOL = 6


def __is_int(value: str) -> bool:
    try:
        # converting to integer
        int(value)
        return True
    except ValueError:
        return False


def _ignorable(last_path: str) -> bool:
    # first case: start with a starter, then there is a number without a splitter
    for name, member in Starter.__members__.items():
        if last_path.upper().startswith(name):
            # is there a number after that?
            if len(last_path) > len(name):
                potential_discnumber: str = last_path[len(name)]
                if __is_int(potential_discnumber):
                    return True
    # second case: start with a starter, then there is a number after a splitter
    for name, member in Starter.__members__.items():
        if last_path.upper().startswith(name):
            splitted: list = split_string(last_path, __split_characters)
            if splitted is not None and len(splitted) >= 2:
                last: str = splitted[1]
                if __is_int(last):
                    return True
    return False


def get_last_path_element(path: str) -> str:
    return os.path.basename(os.path.normpath(path))


def get_dir_from_path(path: str) -> str:
    return os.path.dirname(path)


def get_album_base_path(path: str) -> str:
    last_path = os.path.basename(os.path.normpath(path))
    last_path_ignorable: bool = _ignorable(last_path)
    if last_path_ignorable:
        return os.path.split(path)[0]
    return path


class __Decorated_Song:

    def __init__(self, song: Song):
        self._song: Song = song
        self._disc: int = song.getDiscNumber() if song.getDiscNumber() else 0
        self._track: int = song.getTrack() if song.getTrack() else 0
        path: str = get_dir_from_path(song.getPath())
        last_path = os.path.basename(os.path.normpath(path))
        last_path_ignorable: bool = _ignorable(last_path)
        if last_path_ignorable:
            self._path: str = os.path.split(path)[0]
        else:
            self._path: str = path

    def getSong(self) -> Song: return self._song
    def getPath(self) -> str: return self._path
    def getDisc(self) -> int: return self._disc
    def getTrack(self) -> int: return self._track


def __compare_decorated_song(left: __Decorated_Song, right: __Decorated_Song) -> int:
    cmp: int
    left_album: str = left.getSong().getAlbum() if left.getSong().getAlbum() else ""
    right_album: str = right.getSong().getAlbum() if right.getSong().getAlbum() else ""
    cmp = -1 if left_album < right_album else 0 if left_album == right_album else 1
    if cmp == 0:
        cmp = -1 if left.getPath() < right.getPath() else 0 if left.getPath() == right.getPath() else 1
    if cmp == 0:
        cmp = -1 if left.getDisc() < right.getDisc() else 0 if left.getDisc() == right.getDisc() else 1
    if cmp == 0:
        cmp = -1 if left.getTrack() < right.getTrack() else 0 if left.getTrack() == right.getTrack() else 1
    return cmp


class MultiCodecAlbum(Enum):
    NO = 0
    YES = 1


class SortSongListResult:

    def __init__(
            self,
            codec_set_by_path: dict[str, set[str]],
            song_list: list[Song],
            multi_codec_album: MultiCodecAlbum):
        self._codec_set_by_path: dict[str, set[str]] = codec_set_by_path
        self._song_list: list[Song] = song_list
        self._multi_codec_album: MultiCodecAlbum = multi_codec_album

    def getCodecSetByPath(self) -> set[str]:
        return self._codec_set_by_path

    def getAlbumVersionCount(self) -> int:
        return len(self._codec_set_by_path.keys())

    def getSongList(self) -> list[Song]:
        return copy.deepcopy(self._song_list)

    def getMultiCodecAlbum(self) -> MultiCodecAlbum:
        return self._multi_codec_album


def get_album_release_date(album: Album, item_key=constants.ItemKey.ORIGINAL_RELEASE_DATE) -> ReleaseDate:
    # extract dict
    rd_dict: dict[str, str] = album.getItem().getByName(item_key.value)
    if not rd_dict:
        return None
    # validate
    # year is mandatory
    if constants.DictKey.YEAR.value not in rd_dict:
        return None
    # if day is present, we also need the month
    if constants.DictKey.DAY.value in rd_dict and constants.DictKey.MONTH.value not in rd_dict:
        return None
    # all good, build ReleaseDate
    return ReleaseDate(rd_dict)


def get_formatted_album_release_date(album: Album) -> str:
    return __get_formatted_album_date(album=album, item_key=constants.ItemKey.RELEASE_DATE)


def get_formatted_album_original_release_date(album: Album) -> str:
    return __get_formatted_album_date(album=album, item_key=constants.ItemKey.ORIGINAL_RELEASE_DATE)


class AlbumReleaseDateType(Enum):
    RELEASE_DATE = 1
    ORIGINAL_RELEASE_DATE = 2


def get_formatted_album_date_from_metadata(
        album_metadata: AlbumMetadata,
        release_date_type: AlbumReleaseDateType) -> str:
    model_key_list: list[AlbumMetadataModel] = ([
        AlbumMetadataModel.ALBUM_RELEASE_DATE_YEAR,
        AlbumMetadataModel.ALBUM_RELEASE_DATE_MONTH,
        AlbumMetadataModel.ALBUM_RELEASE_DATE_DAY]
        if release_date_type == AlbumReleaseDateType.RELEASE_DATE
        else [
            AlbumMetadataModel.ALBUM_ORIGINAL_RELEASE_DATE_YEAR,
            AlbumMetadataModel.ALBUM_ORIGINAL_RELEASE_DATE_MONTH,
            AlbumMetadataModel.ALBUM_ORIGINAL_RELEASE_DATE_DAY])
    y: int = album_metadata.get_value(model_key_list[0])
    if y is None:
        return None
    m: int = album_metadata.get_value(model_key_list[1])
    d: int = album_metadata.get_value(model_key_list[2])
    if (m is None or d is None):
        # just use the year
        return str(y)
    # ok to combine
    return f"{y:04}-{m:02}-{d:02}"


def __get_formatted_album_date(album: Album, item_key: constants.ItemKey) -> str:
    release_date: ReleaseDate = ReleaseDate(album.getItem().getByName(item_key.value))
    if release_date is None:
        return None
    # split and return
    y: int = release_date.year
    if y is None:
        return None
    m: int = release_date.month
    d: int = release_date.day
    if (m is None or d is None):
        # just use the year
        return str(y)
    # ok to combine
    return f"{y:04}-{m:02}-{d:02}"


def _get_album_release_year(album: Album, item_key: constants.ItemKey) -> int:
    ord_dict: dict[str, any] = album.getItem().getByName(item_key.value)
    if ord_dict is None:
        return None
    # just return year.
    return ord_dict["year"] if "year" in ord_dict else None


def sort_song_list(song_list: list[Song]) -> SortSongListResult:
    dec_list: list[__Decorated_Song] = []
    codec_dict: dict[str, int] = {}
    multi_codec: MultiCodecAlbum = MultiCodecAlbum.NO
    codec_set_by_path: dict[str, set[str]] = {}
    for song in song_list:
        dec: __Decorated_Song = __Decorated_Song(song)
        dec_list.append(dec)
        if dec.getPath() not in codec_set_by_path:
            codec_set_by_path[dec.getPath()] = set()
        codec_per_path_set: set[str] = codec_set_by_path[dec.getPath()]
        if song.getSuffix().lower() not in codec_per_path_set:
            codec_per_path_set.add(song.getSuffix().lower())
        if song.getSuffix().lower() not in codec_dict:
            codec_dict[song.getSuffix().lower()] = 1
        else:
            codec_dict[song.getSuffix().lower()] = codec_dict[song.getSuffix().lower()] + 1
    multi_codec = MultiCodecAlbum.YES if len(codec_dict) > 1 else MultiCodecAlbum.NO
    dec_list.sort(key=cmp_to_key(__compare_decorated_song))
    result: list[Song] = []
    current: __Decorated_Song
    for current in dec_list:
        result.append(current.getSong())
    return SortSongListResult(
        codec_set_by_path=codec_set_by_path,
        song_list=result,
        multi_codec_album=multi_codec)


def get_album_path_list(album: Album) -> list[str]:
    return get_album_path_list_by_song_list(song_list=album.getSongs())


def get_album_path_list_by_song_list(song_list: list[Song]) -> list[str]:
    song: Song
    path_set: set[str] = set()
    for song in song_list if song_list else []:
        curr_dir: str = get_dir_from_path(song.getPath())
        if curr_dir not in path_set:
            path_set.add(curr_dir)
    return list(path_set)


def get_album_path_list_joined(song_list: list[Song]) -> str:
    # this will return None if the album does not have songs
    # so if the album comes from a album list, the result will always be None
    start: float = time.time()
    result: str = persistence_constants.Separator.PATH.value.join(get_album_path_list_by_song_list(song_list=song_list))
    elapsed: float = time.time() - start
    if config.get_verbose_logging():
        msgproc.log(f"get_album_path_list_joined song_count [{len(song_list) if song_list else 0}] -> "
                    f"[{result}] in [{elapsed:.3f}]")
    return result


class AlbumTracks:

    def __init__(
            self,
            codec_set_by_path: dict[str, set[str]],
            album: Album,
            song_list: list[Song],
            art_uri: str,
            multi_codec_album: MultiCodecAlbum):
        self._codec_set_by_path: dict[str, set[str]] = codec_set_by_path
        self._album: Album = album
        self._song_list: list[Song] = song_list
        self._art_uri: str = art_uri
        self._multi_codec_album: MultiCodecAlbum = multi_codec_album

    def getCodecSetByPath(self) -> dict[str, set[str]]:
        return copy.deepcopy(self._codec_set_by_path)

    def getAlbumVersionCount(self) -> int:
        return len(self._codec_set_by_path.keys())

    def getAlbum(self) -> Album:
        return self._album

    def getSongList(self) -> list[Song]:
        return copy.deepcopy(self._song_list)

    def getArtUri(self) -> str:
        return self._art_uri

    def getMultiCodecAlbum(self) -> MultiCodecAlbum:
        return self._multi_codec_album


def strip_substring(initial_str, pattern):
    result_str = initial_str
    match = re.search(pattern, initial_str)
    if match:
        start = match.start()
        end = match.end()
        left = initial_str[0:start] if start > 0 else ""
        right = initial_str[match.end():] if end < (len(initial_str) - 1) else ""
        result_str = left + right
    return result_str


def to_codec_pattern(codec: str, style: CodecDelimiterStyle):
    return f" \\{style.get_left()}{codec}\\{style.get_right()}"


def strip_codec_from_album(album_title: str, codecs: set[str]) -> str:
    stripped_title: str = album_title
    if len(codecs) == 1:
        # get first and only codec
        codecs_str: str = list(codecs)[0]
        style: CodecDelimiterStyle
        for style in CodecDelimiterStyle:
            codec_pattern: str = to_codec_pattern(
                codecs_str,
                CodecDelimiterStyle.ROUND)
            if not album_title.startswith(codec_pattern):
                stripped_title = strip_substring(album_title, codec_pattern)
    return stripped_title


def has_year(album: Album) -> bool:
    album_year: str = get_album_year_str(album)
    return album_year is not None and len(album_year) > 0


def get_album_year_str(album: Album) -> str:
    year: int = album.getYear()
    rd_year: int = _get_album_release_year(album=album, item_key=constants.ItemKey.RELEASE_DATE)
    ord_year: int = _get_album_release_year(album=album, item_key=constants.ItemKey.ORIGINAL_RELEASE_DATE)
    y_list: list[int] = []
    if year:
        y_list.append(year)
    if rd_year and rd_year not in y_list:
        y_list.append(rd_year)
    if ord_year and ord_year not in y_list:
        y_list.append(ord_year)
    y_list.sort()
    return "/".join(str(x) for x in y_list)


# Tests
a1d1: str = "Disc 1 - Studio Album"
a1d2: str = "Disc 2 Live Album"

if not _ignorable(a1d1):
    raise Exception(f"Ignorable not working properly, [{a1d1}] should be ignorable")
if not _ignorable(a1d2):
    raise Exception(f"Ignorable not working properly, [{a1d2}] should be ignorable")



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime


class TileImage:

    def __init__(self):
        pass

    @property
    def tile_type(self) -> str:
        return self._tile_type

    @tile_type.setter
    def tile_type(self, value: str):
        self._tile_type: str = value

    @property
    def tile_id(self) -> str:
        return self._tile_id

    @tile_id.setter
    def tile_id(self, value: str):
        self._tile_id: str = value

    @property
    def tile_image(self) -> str:
        return self._tile_image

    @tile_image.setter
    def tile_image(self, value: str):
        self._tile_image: str = value

    @property
    def update_time(self) -> datetime.datetime:
        return self._update_time

    @update_time.setter
    def update_time(self, value: datetime.datetime):
        self._update_time: datetime.datetime = value



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from common_data_structures import ArtistIdName
from enum import Enum


class SongArtistType(Enum):
    SONG_ALBUM_ARTIST = 1
    SONG_ARTIST = 2


class SongArtist:

    def __init__(
            self,
            song_id: str,
            artist_id: str,
            artist_name: str):
        self.__song_id: str = song_id
        self.__artist: ArtistIdName = ArtistIdName(
            artist_id=artist_id,
            artist_name=artist_name)

    @property
    def song_id(self) -> str:
        return self.__song_id

    @property
    def artist_id(self) -> str:
        return self.__artist.artist_id

    @property
    def artist_name(self) -> str:
        return self.__artist.artist_name


class SongContributor:

    def __init__(
            self,
            song_id: str,
            role: str,
            sub_role: str,
            artist_id: str,
            artist_name: str):
        self.__song_id: str = song_id
        self.__role: str = role
        self.__sub_role: str = sub_role
        self.__artist: ArtistIdName = ArtistIdName(
            artist_id=artist_id,
            artist_name=artist_name)

    @property
    def song_id(self) -> str:
        return self.__song_id

    @property
    def role(self) -> str:
        return self.__role

    @property
    def sub_role(self) -> str:
        return self.__sub_role

    @property
    def artist_id(self) -> str:
        return self.__artist.artist_id

    @property
    def artist_name(self) -> str:
        return self.__artist.artist_name



# Copyright (C) 2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sqlite3
import sqlhelper


def __sqlite3_execute_update(
        connection: sqlite3.Connection,
        sql: str,
        data: tuple,
        cursor=None,
        do_commit: bool = True) -> int:
    the_cursor = cursor if cursor is not None else connection.cursor()
    the_cursor.execute(sql, data)
    row_count: int = the_cursor.rowcount
    if cursor is None:
        the_cursor.close()
    if do_commit:
        connection.commit()
    return row_count


def __sqlite3_select(
        connection: sqlite3.Connection,
        sql: str,
        parameters: tuple) -> list[any]:
    cursor = connection.cursor()
    cursor.execute(sql, parameters)
    rows = cursor.fetchall()
    cursor.close()
    return rows if rows else []


def get_sqlite3_executor(connection: sqlite3.Connection) -> sqlhelper.SqlExecutor:
    return lambda sql, data, do_commit: __sqlite3_execute_update(
        connection=connection,
        sql=sql,
        data=data,
        do_commit=do_commit)


def get_sqlite3_selector(connection: sqlite3.Connection) -> sqlhelper.SqlSelector:
    return lambda sql, parameters: __sqlite3_select(
        connection=connection,
        sql=sql,
        parameters=parameters)



#!/usr/bin/env python3
#
# A lot of code copied from the Kodi Tidal addon which is:
# Copyright (C) 2014 Thomas Amland
#
# Additional code and modifications:
# Copyright (C) 2016 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import sys
import os
import json
import re
import cmdtalkplugin
from upmplgutils import *
from xbmcplug import *

# Using kodi routing module
from routing import Plugin

plugin = Plugin("")

from session import Session

qobidprefix = "0$qobuz$"

# We maybe should retrieve this from the "index" call (cf notes/qobuz/featured/index/json)
_playlist_tags = (
    ("hi-res", "Hi-Res"),
    ("new", "New"),
    ("focus", "Themes"),
    ("danslecasque", "Artist's choices"),
    ("label", "Labels"),
    ("mood", "Moods"),
    ("artist", "Artists"),
    ("events", "Events"),
    ("auditoriums", "Hi-Fi Audio Partners"),
    ("popular", "Popular"),
)

# Func name to method mapper
dispatcher = cmdtalkplugin.Dispatch()
# Pipe message handler
msgproc = cmdtalkplugin.Processor(dispatcher)

# Initial formatid (mp3). Will be reset on login. This is because we may need a session when
# get_appid is called (when the plugin is used with the credentials service). It seems that a local
# session inside getappid would suffice, TBD check).
session = Session(5)

_g_loginok = False


def maybelogin(a={}):
    global session
    global formatid
    global httphp
    global pathprefix
    global _g_loginok
    global renum_tracks
    global explicit_item_numbers
    global prepend_artist_to_album

    # Do this always
    setidprefix(qobidprefix)

    if _g_loginok:
        return True

    if "UPMPD_HTTPHOSTPORT" not in os.environ:
        raise Exception("No UPMPD_HTTPHOSTPORT in environment")
    httphp = os.environ["UPMPD_HTTPHOSTPORT"]
    if "UPMPD_PATHPREFIX" not in os.environ:
        raise Exception("No UPMPD_PATHPREFIX in environment")
    pathprefix = os.environ["UPMPD_PATHPREFIX"]
    if "UPMPD_CONFIG" not in os.environ:
        raise Exception("No UPMPD_CONFIG in environment")

    # Format id: 5 for MP3 320, 6 for FLAC Lossless, 7 for FLAC
    # Hi-Res 24 bit =< 96kHz, 27 for FLAC Hi-Res 24 bit >96 kHz &
    # =< 192 kHz
    formatid = getOptionValue("qobuzformatid", 6)
    # Decide if we fetch the actual track audio details when listing containers. This is expensive
    # and normally not needed.
    needaudiodetails = getOptionValue("qobuzaudiodetails", False)
    showalbumrateandbits = getOptionValue("qobuzalbrateandbits", False)

    # renum_tracks will instruct the trackentries method to generate track numbers by
    # counting the displayed elements, so the tracks should be always presented in the correct order
    # in the case of playlists and, most importantly, multi-disc albums
    # This is afaik to the benefit of kodi mostly
    renum_tracks = getOptionValue("qobuzrenumtracks", True)

    # explicit_item_numbers will cause the lists to be prepended with a number in
    # square brackets
    # It is disabled by default
    # This is afaik to the benefit of kodi mostly which always tries
    # to sort the lists by name
    explicit_item_numbers = getOptionValue("qobuzexplicititemnumbers", False)

    # prepend_artist_to_album will cause the album list to include the
    # artist before the album title so the contents of the list are more
    # accessible on kodi, which does not display the name of the parent directory
    # This is afaik to the benefit of kodi mostly, for better usability
    prepend_artist_to_album = getOptionValue("qobuzprependartisttoalbum", False)

    # Set default values for track listings
    if formatid == 5:
        setMimeAndSamplerate("audio/mpeg", "44100")
    if formatid == 6:
        setMimeAndSamplerate("application/flac", "44100")
    elif formatid == 7:
        setMimeAndSamplerate("application/flac", "96000", bits="24")
    elif formatid == 27:
        setMimeAndSamplerate("application/flac", "192000", bits="24")
    else:
        setMimeAndSamplerate("audio/mpeg", "44100")

    session = Session(
        formatid, fetch_resource_info=needaudiodetails, show_album_maxaudio=showalbumrateandbits
    )

    _g_loginok = session.login()


# The following two (getappid and login) are not used by the media server, they're for use by the
# OpenHome Credentials service
@dispatcher.record("getappid")
def getappid(a):
    appid = getOptionValue("qobuzappid")
    if appid:
        return {"appid": appid}
    appid = session.get_appid()
    return {"appid": appid}


@dispatcher.record("login")
def login(a):
    maybelogin(a)
    appid, token = session.get_appid_and_token()
    if token is None:
        # login failed. cmdtalk does not like None values
        token = ""
    return {"appid": appid, "token": token}


@dispatcher.record("trackuri")
def trackuri(a):
    msgproc.log(f"trackuri: [{a}]")
    maybelogin()

    if "path" not in a:
        raise Exception("trackuri: no 'path' in args")


    path = a["path"]
    if path.split("/")[2] == "oauth":
        msgproc.log("Qobuz: trackuri: OAuth initialisation")
        if "query" not in a:
            msgproc.log("Oauth init: no query values ??")
            return {}
        #msgproc.log(f"OAuth: got query: {a['query']}")
        querymap = json.loads(a["query"])
        auth_code = None
        try:
            auth_code = querymap["code"]
        except:
            auth_code = querymap["code_autorisation"]
        msgproc.log(f"OAuth: got auth_code: {auth_code}")
        # Use the code  and
        session.init_oauth(auth_code)

        # And return page data to display to the user.
        return {"method": "Respond",
                "media_url": "<html><body style='font-family:system-ui;text-align:center'>"
                "<h3>Upmpdcli qobuz plugin initialisation successful</h3>"
                "<p>You can close this tab and return to your terminal.</p></body></html>"
                }
    

    trackid = trackid_from_urlpath(pathprefix, a)
    media_url = session.get_media_url(trackid)

    return {"media_url": media_url}


def track_list(tracks):
    xbmcplugin.entries += trackentries(
        httphp, pathprefix, xbmcplugin.objid, tracks, generate_track_nums=renum_tracks
    )


@dispatcher.record("browse")
def browse(a):
    global xbmcplugin
    msgproc.log("browse: [%s]" % a)
    if "objid" not in a:
        raise Exception("No objid in args")
    objid = a["objid"]
    bflg = a["flag"] if "flag" in a else "children"

    if re.match(r"0\$qobuz\$", objid) is None:
        raise Exception("bad objid [%s]" % objid)

    xbmcplugin = XbmcPlugin(qobidprefix, objid, routeplugin=plugin)

    maybelogin()

    idpath = objid.replace(qobidprefix, "", 1)
    if bflg == "meta":
        m = re.match(r".*\$(.+)$", idpath)
        if m:
            trackid = m.group(1)
            track = session.get_track(trackid)
            track_list(
                [
                    track,
                ]
            )
    else:
        plugin.run([idpath])
    # msgproc.log("%s" % xbmcplugin.entries)
    encoded = json.dumps(xbmcplugin.entries)
    return {"entries": encoded}


@plugin.route("/")
def root():
    xbmcplugin.add_directory("Discover Catalog", whats_new)
    xbmcplugin.add_directory("Discover Genres", root_genres)
    xbmcplugin.add_directory("Favourites", my_music)


@plugin.route("/root_genres")
def root_genres():
    items = session.get_genres()
    xbmcplugin.view(
        items,
        xbmcplugin.urls_from_id(genre_view, items),
        initial_item_num=1 if explicit_item_numbers else None,
    )


@plugin.route("/genre/<genre_id>")
def genre_view(genre_id):
    endpoint_list = [
        ("New Releases", plugin.url_for(featured_albums, genre_id=genre_id, type="new-releases")),
        (
            "Ideal Discography",
            plugin.url_for(featured_albums, genre_id=genre_id, type="ideal-discography"),
        ),
        ("Qobuzissime", plugin.url_for(featured_albums, genre_id=genre_id, type="qobuzissims")),
        ("Editor Picks", plugin.url_for(featured_albums, genre_id=genre_id, type="editor-picks")),
        ("Press Awards", plugin.url_for(featured_albums, genre_id=genre_id, type="press-awards")),
        ("Qobuz Playlists", plugin.url_for(featured_playlists, genre_id=genre_id, tags="None")),
    ]

    item_num = 1
    for endpoint in endpoint_list:
        xbmcplugin.add_directory(
            endpoint[0], endpoint[1], item_num=item_num if explicit_item_numbers else None
        )
        item_num += 1


@plugin.route("/featured/<genre_id>/<type>")
def featured_albums(genre_id, type):
    items = session.get_featured_albums(genre_id=genre_id, type=type)
    update_album_names(items)
    xbmcplugin.view(
        items,
        xbmcplugin.urls_from_id(album_view, items),
        initial_item_num=1 if explicit_item_numbers else None,
    )


# This used to be /featured/<genre_id>/playlist, but this path can be
# matched by the one for genre_view_type, and the wrong function may
# be called, depending on the rules ordering (meaning we had the
# problem on an rpi, but not ubuntu...)
@plugin.route("/featured_playlists/<genre_id>/<tags>")
def featured_playlists(genre_id, tags):
    items = session.get_featured_playlists(genre_id=genre_id, tags=tags)
    xbmcplugin.view(
        items,
        xbmcplugin.urls_from_id(playlist_view, items),
        initial_item_num=1 if explicit_item_numbers else None,
    )


@plugin.route("/featured_artists")
def featured_artists():
    items = session.get_featured_artists()
    xbmcplugin.view(
        items,
        xbmcplugin.urls_from_id(artist_view, items),
        initial_item_num=1 if explicit_item_numbers else None,
    )


@plugin.route("/whats_new")
def whats_new():
    xbmcplugin.add_directory(
        "Playlists", plugin.url_for(featured_playlists, genre_id=None, tags=None)
    )
    xbmcplugin.add_directory(
        "Albums (ideal discography)",
        plugin.url_for(featured_albums, genre_id=None, type="ideal-discography"),
    )
    xbmcplugin.add_directory(
        "Albums (Qobuzissime)", plugin.url_for(featured_albums, genre_id=None, type="qobuzissims")
    )
    xbmcplugin.add_directory(
        "Albums (new)", plugin.url_for(featured_albums, genre_id=None, type="new-releases-full")
    )
    xbmcplugin.add_directory("Artists", plugin.url_for(featured_artists))

    for tag, label in _playlist_tags:
        xbmcplugin.add_directory(
            f"Playlists ({label})", plugin.url_for(featured_playlists, genre_id=None, tags=tag)
        )


def update_album_names(items):
    for item in items if items and len(items) > 0 else []:
        update_album_name(item)


def update_album_name(item):
    if prepend_artist_to_album:
        artist = item.artist.name if item.artist else None
        if artist and len(artist) > 0:
            item.name = f"{artist} - {item.name}"


@plugin.route("/my_music")
def my_music():
    xbmcplugin.add_directory("Albums", favourite_albums)
    xbmcplugin.add_directory("Tracks", favourite_tracks)
    xbmcplugin.add_directory("Artists", favourite_artists)
    xbmcplugin.add_directory("Playlists", favourite_playlists)
    pass


@plugin.route("/album/<album_id>")
def album_view(album_id):
    track_list(session.get_album_tracks(album_id))


@plugin.route("/playlist/<playlist_id>")
def playlist_view(playlist_id):
    track_list(session.get_playlist_tracks(playlist_id))


@plugin.route("/artist/<artist_id>")
def artist_view(artist_id):
    xbmcplugin.add_directory("Similar Artists", plugin.url_for(similar_artists, artist_id))
    albums = session.get_artist_albums(artist_id)
    xbmcplugin.view(
        albums,
        xbmcplugin.urls_from_id(album_view, albums),
        initial_item_num=1 if explicit_item_numbers else None,
    )


@plugin.route("/artist/<artist_id>/similar")
def similar_artists(artist_id):
    artists = session.get_artist_similar(artist_id)
    xbmcplugin.view(
        artists,
        xbmcplugin.urls_from_id(artist_view, artists),
        initial_item_num=1 if explicit_item_numbers else None,
    )


@plugin.route("/favourite_tracks")
def favourite_tracks():
    track_list(session.user.favorites.tracks())


@plugin.route("/favourite_artists")
def favourite_artists():
    try:
        items = session.user.favorites.artists()
    except Exception as err:
        msgproc.log("session.user.favorite.artists failed: %s" % err)
        return
    if items:
        msgproc.log("First artist name %s" % items[0].name)
        xbmcplugin.view(
            items,
            xbmcplugin.urls_from_id(artist_view, items),
            initial_item_num=1 if explicit_item_numbers else None,
        )


@plugin.route("/favourite_albums")
def favourite_albums():
    items = session.user.favorites.albums()
    update_album_names(items)
    xbmcplugin.view(
        items,
        xbmcplugin.urls_from_id(album_view, items),
        initial_item_num=1 if explicit_item_numbers else None,
    )


@plugin.route("/favourite_playlists")
def favourite_playlists():
    items = session.user.favorites.playlists()
    xbmcplugin.view(
        items,
        xbmcplugin.urls_from_id(playlist_view, items),
        initial_item_num=1 if explicit_item_numbers else None,
    )


@dispatcher.record("search")
def search(a):
    global xbmcplugin
    msgproc.log("search: [%s]" % a)
    objid = a["objid"]
    field = a["field"] if "field" in a else None
    value = a["value"]
    objkind = a["objkind"] if "objkind" in a and a["objkind"] else None

    if re.match(r"0\$qobuz\$", objid) is None:
        raise Exception("bad objid [%s]" % objid)

    xbmcplugin = XbmcPlugin(qobidprefix, objid, routeplugin=plugin)

    maybelogin()

    if field and field not in ["artist", "album", "playlist", "track", "title"]:
        msgproc.log("Unknown field '%s'" % field)
        field = "title"
    elif field == "track":
        # plgwithslave.cpp bug: uses 'track' where it should use 'title'
        # (field title is valid for all kinds of objects).
        field = "title"

    if objkind and objkind not in ["track", "artist", "album", "playlist"]:
        msgproc.log("Unknown objkind '%s'" % objkind)
        objkind = "track"

    # Qobuz types are plural: 'tracks', 'albums', 'artists' or 'playlists'
    qkind = objkind + "s" if objkind else None
    searchresults = session.search(field, value, qkind)

    # Display the requested results: maybe all kinds or one specific kind.
    if objkind is None or objkind == "artist":
        xbmcplugin.view(
            searchresults.artists,
            xbmcplugin.urls_from_id(artist_view, searchresults.artists),
            end=False,
        )
    if objkind is None or objkind == "album":
        xbmcplugin.view(
            searchresults.albums,
            xbmcplugin.urls_from_id(album_view, searchresults.albums),
            end=False,
        )
    if objkind is None or objkind == "playlist":
        xbmcplugin.view(
            searchresults.playlists,
            xbmcplugin.urls_from_id(playlist_view, searchresults.playlists),
            end=False,
        )
    if objkind is None or objkind == "track":
        track_list(searchresults.tracks)

    # msgproc.log("%s" % xbmcplugin.entries)
    encoded = json.dumps(xbmcplugin.entries)
    return {"entries": encoded}


msgproc.log("Qobuz running")
maybelogin()
msgproc.mainloop()



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class CodecDelimiterStyle(Enum):

    SQUARE = 0, "[", "]"
    ROUND = 1, "(", ")"

    def __init__(self, num: int, left: str, right: str):
        self.__num: int = num
        self.__left: str = left
        self.__right: str = right

    def get_num(self) -> int: return self.__num
    def get_left(self) -> str: return self.__left
    def get_right(self) -> str: return self.__right



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class PlayedTrackRequest:

    def __init__(self):
        self._track_id: str = None
        self._album_id: str = None
        self._album_track_count: int = None
        self._track_name: str = None
        self._track_duration: int = None
        self._track_num: int = None
        self._volume_num: int = None
        self._album_num_volumes: int = None
        self._album_name: str = None
        self._audio_quality: str = None
        self._album_artist_name: str = None
        self._image_url: str = None
        self._explicit: int = None
        self._artist_name: str = None
        self._album_duration: int = None
        self._bit_depth: int = None
        self._sample_rate: int = None

    @property
    def track_id(self) -> str:
        return self._track_id

    @track_id.setter
    def track_id(self, value: str):
        self._track_id: str = value

    @property
    def album_id(self) -> str:
        return self._album_id

    @album_id.setter
    def album_id(self, value: str):
        self._album_id: str = value

    @property
    def album_track_count(self) -> int:
        return self._album_track_count

    @album_track_count.setter
    def album_track_count(self, value: str):
        self._album_track_count: int = value

    @property
    def track_name(self) -> str:
        return self._track_name

    @track_name.setter
    def track_name(self, value: str):
        self._track_name: str = value

    @property
    def track_duration(self) -> int:
        return self._track_duration

    @track_duration.setter
    def track_duration(self, value: int):
        self._track_duration: int = value

    @property
    def track_num(self) -> int:
        return self._track_num

    @track_num.setter
    def track_num(self, value: int):
        self._track_num: int = value

    @property
    def volume_num(self) -> int:
        return self._volume_num

    @volume_num.setter
    def volume_num(self, value: int):
        self._volume_num: int = value

    @property
    def album_num_volumes(self) -> int:
        return self._album_num_volumes

    @album_num_volumes.setter
    def album_num_volumes(self, value: int):
        self._album_num_volumes: int = value

    @property
    def album_name(self) -> str:
        return self._album_name

    @album_name.setter
    def album_name(self, value: str):
        self._album_name: str = value

    @property
    def audio_quality(self) -> str:
        return self._audio_quality

    @audio_quality.setter
    def audio_quality(self, value: str):
        self._audio_quality: str = value

    @property
    def album_artist_name(self) -> str:
        return self._album_artist_name

    @album_artist_name.setter
    def album_artist_name(self, value: str):
        self._album_artist_name: str = value

    @property
    def image_url(self) -> str:
        return self._image_url

    @image_url.setter
    def image_url(self, value: str):
        self._image_url: str = value

    @property
    def explicit(self) -> int:
        return self._explicit

    @explicit.setter
    def explicit(self, value: int):
        self._explicit: int = value

    @property
    def artist_name(self) -> str:
        return self._artist_name

    @artist_name.setter
    def artist_name(self, value: str):
        self._artist_name: str = value

    @property
    def album_duration(self) -> int:
        return self._album_duration

    @album_duration.setter
    def album_duration(self, value: int):
        self._album_duration: int = value

    @property
    def bit_depth(self) -> int:
        return self._bit_depth

    @bit_depth.setter
    def bit_depth(self, value: int):
        self._bit_depth: int = value

    @property
    def sample_rate(self) -> int:
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, value: int):
        self._sample_rate: int = value



#!/usr/bin/env python3
#
# Copyright (C) 2021 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import json
import re
import cmdtalkplugin
import conftree

from upmplgutils import uplog, setidprefix,getConfigObject
import upradioconf

# Prefix for object Ids. This must be consistent with what contentdirectory.cxx does
_g_myprefix = "0$upradios$"
setidprefix("upradios")

# Func name to method mapper
dispatcher = cmdtalkplugin.Dispatch()
# Pipe message handler
msgproc = cmdtalkplugin.Processor(dispatcher)

def _initradios():
    global _g_radios
    _g_radios = upradioconf.UpmpdcliRadios(getConfigObject())
    
@dispatcher.record('trackuri')
def trackuri(a):
    # We generate URIs which directly point to the radio, so this
    # method should never be called.
    raise Exception("trackuri: should not be called for upradios!")


@dispatcher.record('browse')
def browse(a):
    #msgproc.log(f"browse: args: --{a}--")
    if 'objid' not in a:
        raise Exception("No objid in args")
    # Note: objid should either be our root or a radio id, which we don't check.
    objid = a['objid']
    
    entries = []
    bflg = a['flag'] if 'flag' in a else 'children'
    if bflg == 'meta':
        try:
            pid = _g_myprefix
            idx = upradioconf.radioIndexFromId(objid)
            entries.append(upradioconf.radioToEntry(pid, objid, _g_radios.get_radio(idx)))
        except:
            pass
    else:
        for radio in _g_radios:
            entries.append(upradioconf.radioToEntry(objid, None, radio))

    encoded = json.dumps(entries)
    # msgproc.log(f"browse: returning --{entries}--")
    return {"entries" : encoded, "nocache" : "0"}


@dispatcher.record('search')
def search(a):
    msgproc.log("search: [%s]" % a)
    return {"entries" : "[]", "nocache" : "0"}


_initradios()

msgproc.mainloop()
